--
-- Script to create instantLicense and its tables
--
set statement_timeout = 0;
set client_encoding = 'UTF8';
set standard_conforming_strings = on;
set check_function_bodies = false;
set client_min_messages = warning;

--
-- TOC entry 3007 (class 1262 OID 26325)
-- Name: instantLicense; Type: DATABASE; Schema: -; Owner: postgres
--

create database "instantLicense" with template = template1 encoding = '{{ pgsql_locale_encoding }}' lc_collate = '{{ pgsql_lc_collate }}' lc_ctype = '{{ pgsql_lc_ctype }}';


alter database "instantLicense" owner to i3dhub;

\connect "instantLicense"

set statement_timeout = 0;
set client_encoding = 'UTF8';
set standard_conforming_strings = on;
set check_function_bodies = false;
set client_min_messages = warning;

create schema i3dh;

alter schema i3dh owner to i3dhub;

-- Drop table

-- DROP TABLE i3dh.license

create table i3dh.license
(
    id         serial  not null,
    license_id varchar null,
    product_id varchar null,
    constraint license_pkey primary key (id)
)
    with (
        oids= false
    );

-- Permissions

alter table i3dh.license
    owner to i3dhub;


-- Drop table

-- DROP TABLE i3dh.checked_features

create table i3dh.checked_features
(
    feature_name varchar null,
    license_id   int4    null,
    id           serial  not null
)
    with (
        oids= false
    );

-- Permissions

alter table i3dh.checked_features
    owner to i3dhub;


-- Drop table

-- DROP TABLE i3dh.grants

create table i3dh.grants
(
    checkout_time timestamptz null,
    checkin_time  timestamptz null,
    feature_id    int4        null,
    id            serial      not null,
    grant_id      bytea[]     null,
    expires_at    timestamptz null,
    service_id    varchar     null
)
    with (
        oids= false
    );

-- Permissions

alter table i3dh.grants
    owner to i3dhub;


-- Add missing primary key for foreign key constraint used in the usage table
-- Otherwise an error would occur
alter table i3dh.grants
    add primary key (id);


-- Drop table

-- DROP TABLE i3dh.usages

create table i3dh.usages
(
    id         serial primary key,
    start_time timestamptz null,
    end_time   timestamptz null,
    grant_id   bigint
        constraint fk_grant_id references i3dh.grants (id) on delete cascade
)
    with (
        oids= false
    );

-- Permissions

alter table i3dh.usages
    owner to i3dhub;

-- Drop table

-- DROP TABLE i3dh.sessions

create table i3dh.session_features
(
    id              serial primary key,
    cluster_id      varchar     not null,
    user_id         varchar     not null,
    app_instance_id varchar     not null,
    feature_mask    bigint      not null,
    time            timestamptz not null
)
    with (
        oids= false
    );

-- Permissions

alter table i3dh.session_features
    owner to i3dhub;

-- Drop table

-- drop table i3dh.summary

create table i3dh.summary
(
    id              serial primary key,
    cluster_id      varchar     not null,
    user_id         varchar     not null,
    app_instance_id varchar     not null,
    features        bigint      not null,
    costs           bigint      not null,
    start_time      timestamptz not null,
    duration        interval    not null
)
    with (
        oids= false
    );

-- Permissions

alter table i3dh.summary
    owner to i3dhub;

-- Drop materialized view

-- drop materialized view i3dh.summary_view

create materialized view i3dh.summary_view as
select *
from i3dh.summary;

-- Permissions

alter materialized view i3dh.summary_view owner to i3dhub;

-- Drop table

-- DROP TABLE i3dh.schema_migrations;

CREATE TABLE i3dh.schema_migrations
(
    "version" int8 NOT NULL,
    dirty     bool NOT NULL,
    CONSTRAINT schema_migrations_pkey PRIMARY KEY (version)
);

-- Permissions

alter table i3dh.schema_migrations
    owner to i3dhub;
