
-- search for extensions in l3d schema
SET search_path = l3d, extensions;

-- Function: del_cache_by_id(uuid, bigint, bigint)

DROP FUNCTION IF EXISTS l3d.del_cache_by_id(character varying, bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.del_cache_by_id(
    c_uuid character varying,
    c_id0 bigint,
    c_id1 bigint)
  RETURNS void
  AS $BODY$
	BEGIN
    EXECUTE $$
    DELETE FROM l3d.cache
    WHERE l3d.cache.uuid = $1
    AND ext_cache_id0 = $2
    AND ext_cache_id1 = $3
    $$ USING C_UUID, C_ID0, C_ID1;
    END;
  $BODY$
  LANGUAGE plpgsql VOLATILE
  COST 100;
ALTER FUNCTION l3d.del_cache_by_id(character varying, bigint, bigint)
  OWNER TO i3dhub;




-- FUNCTION: l3d.del_cache_by_uuid(uuid)

DROP FUNCTION IF EXISTS l3d.del_cache_by_uuid(character varying);

CREATE OR REPLACE FUNCTION l3d.del_cache_by_uuid(
	c_uuid character varying)
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$

    BEGIN
    	PERFORM l3d.su_del_cache_by_ids( ARRAY(
        SELECT id FROM l3d.cache WHERE uuid = c_uuid));
    END;


$BODY$;

ALTER FUNCTION l3d.del_cache_by_uuid(character varying)
    OWNER TO i3dhub;




-- Function: ins_auxnode(bigint, bigint, bigint)

DROP FUNCTION IF EXISTS l3d.ins_auxnode(bigint, bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_auxnode(
    p_auxnodetype_id bigint,
    p_cache_id bigint,
    p_auxnode_id bigint)
  RETURNS bigint
  	LANGUAGE 'plpgsql'
    VOLATILE
  	COST 100
  AS $BODY$
		DECLARE id_val bigint;
        BEGIN
        EXECUTE $$
        INSERT INTO l3d.aux_node(aux_node_type_id,cache_id,aux_node_id)
        VALUES ($1, $2, $3)
        RETURNING id
        $$ INTO id_val USING p_AUXNodeType_ID, p_Cache_ID, p_AUXNode_ID ;
        RETURN id_val;
        END;
	$BODY$;
ALTER FUNCTION l3d.ins_auxnode(bigint, bigint, bigint)
  OWNER TO i3dhub;



-- Function: ins_auxnodetype(character varying)

DROP FUNCTION IF EXISTS l3d.ins_auxnodetype(character varying);

CREATE OR REPLACE FUNCTION l3d.ins_auxnodetype(p_name character varying)
  RETURNS bigint
  LANGUAGE 'plpgsql'
  VOLATILE
  COST 100
  AS $BODY$
	DECLARE id_val bigint;
    BEGIN
    EXECUTE $$
    INSERT INTO l3d.aux_node_type(name)
    VALUES ($1)
    RETURNING id
    $$ INTO id_val USING  p_Name;
    RETURN id_val;
    END;
  $BODY$;
ALTER FUNCTION l3d.ins_auxnodetype(character varying)
  OWNER TO i3dhub;




-- Function: ins_auxvalue(bigint, character varying, bigint)

DROP FUNCTION IF EXISTS l3d.ins_auxvalue(bigint, character varying, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_auxvalue(
    p_AUXValueName_id bigint,
    p_value character varying,
    p_auxnode_id bigint)
  RETURNS bigint
  LANGUAGE 'plpgsql'
  VOLATILE
  COST 100AS
	$BODY$
		DECLARE id_val bigint;
        BEGIN
        	EXECUTE $$
        	 	INSERT INTO l3d.aux_value(aux_value_name_id,value_str,aux_node_id)
        	 	VALUES ($1,$2,$3)
        	 	RETURNING id
        	$$ INTO id_val USING p_AUXValueName_ID, p_Value, p_AUXNode_ID;
        RETURN id_val;
        END;
    $BODY$;
ALTER FUNCTION l3d.ins_auxvalue(bigint, character varying, bigint)
  OWNER TO i3dhub;



-- Function: ins_AUXValueName(bigint, character varying)

DROP FUNCTION IF EXISTS l3d.ins_AUXValueName(bigint, character varying);

CREATE OR REPLACE FUNCTION l3d.ins_AUXValueName(
    p_aux_node_type_id bigint,
    p_name character varying)
  RETURNS bigint
  LANGUAGE 'plpgsql'
  VOLATILE
  COST 100
  AS $BODY$
	DECLARE id_val bigint;
    BEGIN
    	EXECUTE $$
    	INSERT INTO l3d.aux_value_name(aux_node_type_id,name)
    	VALUES ($1, $2)
    	RETURNING id
        $$ INTO id_val USING p_aux_node_type_id, p_Name;
    	RETURN id_val;
    END;
$BODY$;

ALTER FUNCTION l3d.ins_AUXValueName(bigint, character varying)
  OWNER TO i3dhub;





-- FUNCTION: l3d.ins_cache(integer, integer, uuid)

DROP FUNCTION IF EXISTS l3d.ins_cache(integer, integer, character varying);

CREATE OR REPLACE FUNCTION l3d.ins_cache(
	p_cacheid0 integer,
	p_cacheid1 integer,
	p_uuid character varying)
RETURNS bigint
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$
	DECLARE
    	id_val bigint;
  BEGIN
  	EXECUTE $$
    	INSERT INTO l3d.cache(ext_cache_id0,ext_cache_id1,uuid)
    	VALUES ($1, $2, $3)
  		RETURNING id
  	$$ INTO id_val USING p_cacheid0, p_cacheid1, p_uuid;
  	RETURN id_val;
  END;
$BODY$;

ALTER FUNCTION l3d.ins_cache(integer, integer, character varying)
    OWNER TO i3dhub;




-- Function: ins_edge(bigint, bigint, bigint)

DROP FUNCTION IF EXISTS l3d.ins_edge(bigint, bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_edge(
    p_edge_id bigint,
    p_shape_id bigint,
    p_cache_id bigint)
  RETURNS bigint
  	LANGUAGE 'plpgsql'
    VOLATILE
    COST 100
  AS $BODY$
	DECLARE id_val bigint;
	BEGIN
    	EXECUTE $$
    	INSERT INTO l3d.edge(ext_edge_id,shape_id,cache_id)
    	VALUES ($1,$2,$3) RETURNING id
        $$
    	INTO id_val USING p_Edge_ID, p_Shape_ID, p_Cache_ID;
    	RETURN id_val;
    END;
$BODY$;

ALTER FUNCTION l3d.ins_edge(bigint, bigint, bigint)
  OWNER TO i3dhub;




-- Function: ins_edgeinstance(bigint, bigint, bigint)

DROP FUNCTION IF EXISTS l3d.ins_edgeinstance(bigint, bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_edgeinstance(
    p_edge_id bigint,
    p_cache_id bigint,
    p_shapeinstance_id bigint)
  RETURNS bigint
  	LANGUAGE 'plpgsql'
    VOLATILE
  	COST 100
  AS $BODY$
	DECLARE id_val bigint;
    BEGIN
    	EXECUTE $$
        INSERT INTO l3d.edge_instance(edge_id,cache_id,shape_instance_id)
    	VALUES ($1,$2,$3)
    	RETURNING id
        $$
    	INTO id_val USING p_Edge_ID, p_Cache_ID, p_ShapeInstance_ID;
    	RETURN id_val;
    END;
$BODY$;

ALTER FUNCTION l3d.ins_edgeinstance(bigint, bigint, bigint)
  OWNER TO i3dhub;



-- Function: ins_edgeinstance_auxnode(bigint)

DROP FUNCTION IF EXISTS l3d.ins_edgeinstance_auxnode(bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_edgeinstance_auxnode(p_edgeinstance_id bigint, p_auxnode_id bigint)
  RETURNS bigint
  LANGUAGE 'plpgsql'
  VOLATILE
  COST 100
 AS $BODY$
	DECLARE id_val bigint;
    BEGIN
    EXECUTE $$
    	INSERT INTO l3d.edge_instance_to_aux_node(edge_instance_id,aux_node_id)
    	VALUES ($1,$2)
    	RETURNING id
    $$
    INTO id_val USING p_EdgeInstance_ID, p_Auxnode_ID;
    RETURN id_val;
    END;
$BODY$;

ALTER FUNCTION l3d.ins_edgeinstance_auxnode(bigint,bigint)
  OWNER TO i3dhub;

-- Function: ins_face(bigint, bigint, bigint)

DROP FUNCTION IF EXISTS l3d.ins_face(bigint, bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_face(
    p_shape_id bigint,
    p_face_id bigint,
    p_cache_id bigint)
  RETURNS bigint
  LANGUAGE plpgsql VOLATILE
  COST 100
  AS $BODY$
		DECLARE id_val bigint;
        BEGIN
        EXECUTE $$
        INSERT INTO l3d.face(shape_id,ext_face_id,cache_id)
        VALUES ($1, $2,$3)
        RETURNING id
        $$ INTO id_val USING p_Shape_ID, p_Face_ID, p_Cache_ID;
        RETURN id_val;
        END;
	$BODY$;

ALTER FUNCTION l3d.ins_face(bigint, bigint, bigint)
  OWNER TO i3dhub;



-- Function: ins_faceinstance(bigint, bigint, bigint)

DROP FUNCTION IF EXISTS l3d.ins_faceinstance(bigint, bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_faceinstance(
    p_face_id bigint,
    p_cache_id bigint,
    p_shapeinstance_id bigint)
  RETURNS bigint
  LANGUAGE plpgsql VOLATILE
  COST 100
  AS $BODY$
	DECLARE id_val bigint;
    BEGIN
    EXECUTE $$
    	INSERT INTO l3d.face_instance(face_id,cache_id,shape_instance_id)
    	VALUES ($1,$2,$3)
    	RETURNING id
    $$ INTO id_val USING p_Face_ID, p_Cache_ID, p_ShapeInstance_ID;
    RETURN id_val;
    END;
  $BODY$;

ALTER FUNCTION l3d.ins_faceinstance(bigint, bigint, bigint)
  OWNER TO i3dhub;



-- Function: ins_faceinstance_auxnode(bigint)

DROP FUNCTION IF EXISTS l3d.ins_faceinstance_auxnode(bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_faceinstance_auxnode(p_faceinstance_id bigint, p_auxnode_id bigint)
  RETURNS bigint
  LANGUAGE 'plpgsql'
  VOLATILE
  COST 100
  AS $BODY$
	DECLARE id_val bigint;
    BEGIN
    	EXECUTE $$
    		INSERT INTO "l3d".l3d.face_instance_to_aux_node(face_instance_id,aux_node_id)
    		VALUES ($1, $2)
   			RETURNING id
        $$ INTO id_val USING p_FaceInstance_ID, p_AuxNode_ID;
    	RETURN id_val;
    END;
$BODY$;
ALTER FUNCTION l3d.ins_faceinstance_auxnode(bigint,bigint)
  OWNER TO i3dhub;



-- Function: ins_node(bigint, bigint, character varying)

DROP FUNCTION IF EXISTS l3d.ins_node(bigint, bigint, character varying);

CREATE OR REPLACE FUNCTION l3d.ins_node(
    p_node_id bigint,
    p_cache_id bigint,
    p_label character varying)
  RETURNS bigint
  LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
  AS $BODY$
	DECLARE id_val bigint;
	BEGIN
	EXECUTE $$
    	INSERT INTO "l3d".l3d.node(ext_node_id,cache_id,"Label")
		VALUES ($1,$2,$3) RETURNING id
        $$
        INTO id_val USING p_node_id,p_cache_id,p_label;
	RETURN id_val;
	END;
 $BODY$;

ALTER FUNCTION l3d.ins_node(bigint, bigint, character varying)
  OWNER TO i3dhub;



-- FUNCTION: l3d.ins_nodes(bigint, bigint[], text[])

DROP FUNCTION IF EXISTS l3d.ins_nodes(bigint, bigint[], text[]);

CREATE OR REPLACE FUNCTION l3d.ins_nodes(
	p_cache_id bigint,
	p_nodes bigint[],
	p_labels text[])
RETURNS boolean
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$

  DECLARE res boolean;
  BEGIN
  IF (array_length(p_nodes,1) = array_length(p_labels,1)) THEN
  	res := true;
    EXECUTE $$
  	INSERT INTO "l3d".l3d.node(ext_node_id,cache_id,"Label")
  	SELECT UNNEST($1),$2,UNNEST($3)
    $$ USING p_Nodes, p_Cache_ID, p_labels;
  ELSE
  	res :=false;
  END IF;

  RETURN res;
  END;

$BODY$;

ALTER FUNCTION l3d.ins_nodes(bigint, bigint[], text[])
    OWNER TO i3dhub;




-- Function: ins_shape(bigint, bigint)

DROP FUNCTION IF EXISTS l3d.ins_shape(bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_shape(
    p_cache_id bigint,
    p_shape_id bigint)
  RETURNS bigint
  LANGUAGE 'plpgsql'
  VOLATILE
  COST 100
 AS $BODY$
	DECLARE id_val bigint;
    BEGIN
    	EXECUTE $$
    	INSERT INTO l3d.shape(cache_id,shape_id)
    	VALUES ($1,$2)
    	RETURNING id
        $$ INTO id_val USING p_Cache_ID, p_Shape_ID;
    	RETURN id_val;
    END;
$BODY$;

ALTER FUNCTION l3d.ins_shape(bigint, bigint)
  OWNER TO i3dhub;



-- Function: ins_shapeinstance(bigint, bigint, bigint, bigint)

DROP FUNCTION IF EXISTS l3d.ins_shapeinstance(bigint, bigint, bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_shapeinstance(
    p_shapeinstance_id bigint,
    p_cache_id bigint,
    p_shape_id bigint,
    p_node_id bigint)
  RETURNS bigint
  LANGUAGE 'plpgsql'
  VOLATILE
  COST 100
  AS $BODY$
  DECLARE id_val bigint;
  BEGIN
  	EXECUTE $$
  	INSERT INTO "l3d"."ir3d_ShapeInstance"(shape_instance_id,cache_id,shape_id,node_id)
  	VALUES ($1,$2,$3,$4)
  	RETURNING id
    $$ INTO id_val USING p_ShapeInstance_ID, p_Cache_ID,p_Shape_ID,p_Node_ID ;
  	RETURN id_val;
  END;
  $BODY$;

ALTER FUNCTION l3d.ins_shapeinstance(bigint, bigint, bigint, bigint)
  OWNER TO i3dhub;



-- FUNCTION: l3d.ins_shapes(bigint, bigint[])

DROP FUNCTION IF EXISTS l3d.ins_shapes(bigint, bigint[]);

CREATE OR REPLACE FUNCTION l3d.ins_shapes(
	p_cache_id bigint,
	p_shapes bigint[])
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$
 BEGIN
 EXECUTE $$
 INSERT INTO l3d.shape(cache_id,shape_id)
 SELECT $1,UNNEST($2)
 $$ USING p_Cache_ID, p_Shapes;
 END;
$BODY$;

ALTER FUNCTION l3d.ins_shapes(bigint, bigint[])
    OWNER TO i3dhub;





-- FUNCTION: l3d.ins_auxvalues(bigint[], character varying[], bigint)

DROP FUNCTION IF EXISTS l3d.ins_auxvalues(bigint[], character varying[], bigint);

CREATE OR REPLACE FUNCTION l3d.ins_auxvalues(
	p_AUXValueNames bigint[],
	p_values character varying[],
	p_auxnode_id bigint)
RETURNS bigint
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$

  DECLARE res boolean;
  BEGIN
  IF array_length(p_values,1) = array_length(p_AUXValueNames,1) THEN
  res := true;
  EXECUTE $$
  	INSERT INTO "l3d".l3d.aux_value(aux_value_name_id,value_str,aux_node_id)
  	SELECT UNNEST($1),UNNEST($2),$3
  $$ USING p_AUXValueNames, p_values, p_auxnode_id;
  ELSE
  res :=false;
  END IF;
  RETURN res;
  END;

$BODY$;

ALTER FUNCTION l3d.ins_auxvalues(bigint[], character varying[], bigint)
    OWNER TO i3dhub;




-- FUNCTION: l3d.ins_auxnodes(bigint[], bigint, bigint[])

DROP FUNCTION IF EXISTS l3d.ins_auxnodes(bigint[], bigint, bigint[]);

CREATE OR REPLACE FUNCTION l3d.ins_auxnodes(
	p_auxnodetypes bigint[],
	p_cache_id bigint,
	p_auxnode_ids bigint[])
RETURNS boolean
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$

  DECLARE res boolean;

  BEGIN
  	IF array_length(p_auxnodetypes,1) = array_length(p_auxnode_ids,1) THEN
  		res :=true;
        EXECUTE $$
  			INSERT INTO l3d.aux_node(aux_node_type_id,cache_id,aux_node_id)
  			SELECT UNNEST(p_AUXNodeTypes),p_Cache_ID,UNNEST(p_AUXNode_IDs)
        $$ USING p_AUXNodeTypes, p_Cache_ID, p_AUXNode_IDs;
  	ELSE
  		res := false;
  	END IF;
  	RETURN res;
  END;

$BODY$;

ALTER FUNCTION l3d.ins_auxnodes(bigint[], bigint, bigint[])
    OWNER TO i3dhub;



-- FUNCTION: l3d.get_cache(character varying, integer, integer)

DROP FUNCTION IF EXISTS l3d.get_cache(character varying, integer, integer);

CREATE OR REPLACE FUNCTION l3d.get_cache(
	p_uuid character varying,
	p_cid0 integer,
	p_cid1 integer)
RETURNS bigint
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$
	DECLARE c_id bigint;
BEGIN
	EXECUTE $$
    SELECT * FROM l3d.cache
	WHERE uuid = $1
    AND ext_cache_id0 = $2
    AND ext_cache_id1 = $3
	AND active=true
    $$ INTO c_id USING p_uuid, p_cid0, p_cid1;
    RETURN c_id;
    END;
$BODY$;

ALTER FUNCTION l3d.get_cache(character varying, integer, integer)
    OWNER TO i3dhub;


-- FUNCTION: l3d.get_caches(character varying[])

DROP FUNCTION IF EXISTS l3d.get_caches_ids(cached hstore);
CREATE OR REPLACE FUNCTION l3d.get_caches_ids(
	caches hstore)
RETURNS hstore
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$

	DECLARE c_ids hstore;
BEGIN

    SELECT hstore (array_agg(id::text), array_agg(caches->(uuid ||'/'|| ext_cache_id0::varchar ||'/'||  ext_cache_id1::varchar)))   FROM l3d.cache
	WHERE (uuid ||'/'|| ext_cache_id0::varchar ||'/'||  ext_cache_id1::varchar) = ANY( akeys(caches))
	AND active=true
    INTO c_ids;
    RETURN c_ids;
    END;

$BODY$;

ALTER FUNCTION l3d.get_caches_ids(cached hstore)
    OWNER TO i3dhub;



-- FUNCTION: l3d.ins_edge(bigint, bigint, bigint)

DROP FUNCTION IF EXISTS l3d.ins_edges(bigint, bigint, bigint);

CREATE OR REPLACE FUNCTION l3d.ins_edges(
	p_edges_num bigint,
	p_shape_id bigint,
	p_cache_id bigint)
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$
	DECLARE id_val bigint;
	BEGIN
    	EXECUTE $$
    	INSERT INTO l3d.edge(ext_edge_id,shape_id,cache_id)
    	SELECT s, $2 , $3
        FROM  generate_series(1,$1) AS s
        $$
    	USING p_Edges_NUM, p_Shape_ID, p_Cache_ID;
    END;

$BODY$;

ALTER FUNCTION l3d.ins_edges(bigint, bigint, bigint)
    OWNER TO i3dhub;



-- FUNCTION: l3d.su_del_cache_by_id(bigint)

DROP FUNCTION IF EXISTS l3d.su_del_cache_by_id(bigint);

CREATE OR REPLACE FUNCTION l3d.su_del_cache_by_id(
	c_id bigint)
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$

BEGIN

DELETE FROM l3d.index_cahe
        WHERE cache_id=c_id;

DELETE FROM l3d.index_properties
        WHERE cache_id=c_id;

DELETE FROM l3d.index_source
        WHERE cache_id=c_id;

DELETE FROM l3d.index_stats
        WHERE cache_id=c_id;

DELETE FROM l3d.shape
        WHERE cache_id=c_id;

DELETE FROM l3d.node
        WHERE cache_id = c_id;
DELETE FROM l3d.aux_value WHERE aux_node_id IN (
SELECT id FROM l3d.aux_node WHERE cache_id = c_id);

DELETE FROM l3d.aux_node WHERE cache_id = c_id;

DELETE FROM l3d.edge WHERE cache_id = c_id;
DELETE FROM l3d.edge_instance WHERE cache_id = c_id;
DELETE FROM l3d.face WHERE cache_id = c_id;
DELETE FROM l3d.face_instance WHERE cache_id = c_id;
DELETE FROM l3d.shape WHERE cache_id = c_id;
DELETE FROM l3d.shape_instance WHERE cache_id = c_id;

DELETE FROM l3d.cache
        WHERE id=c_id;

END;

$BODY$;

ALTER FUNCTION l3d.su_del_cache_by_id(bigint)
    OWNER TO i3dhub;





-- FUNCTION: l3d.get_aux_by_type(bigint, character varying[], character varying[])

DROP FUNCTION IF EXISTS l3d.get_aux_by_type(bigint, character varying[], integer, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_by_type(
	c_id bigint,
	auxt_types character varying[],
	off_set integer,
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$


BEGIN
	OPEN ref FOR
    SELECT aux_node_id + off_set as "nodeId"
    FROM l3d.aux_node as an, l3d.aux_node_type as ant
	WHERE an.aux_node_type_id = ant.id
    AND cache_id = $1
    AND ant.name = ANY($2::text[]);

    RETURN ref;
    END;

$BODY$;

ALTER FUNCTION l3d.get_aux_by_type(bigint, character varying[], integer, refcursor)
    OWNER TO i3dhub;



DROP FUNCTION IF EXISTS l3d.get_aux_stats(bigint, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_stats(
	cid bigint, ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$

BEGIN
OPEN ref FOR SELECT ant.name as name, COUNT (an.id) as count
FROM l3d.aux_node as an, l3d.aux_node_type as ant
WHERE an.aux_node_type_id = ant.id
AND an.cache_id = cid
GROUP BY ant.name;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_aux_stats(bigint, refcursor)
    OWNER TO i3dhub;





-- FUNCTION: l3d.get_aux_type(bigint, bigint, refcursor)

DROP FUNCTION IF EXISTS l3d.get_aux_type(bigint, bigint, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_type(
	c_id bigint,
	aux_id bigint,
	ref refcursor)
    RETURNS refcursor
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE

AS $BODY$


BEGIN
OPEN ref FOR SELECT ant.name
	FROM l3d.aux_node as an, l3d.aux_node_type as ant
	WHERE an.aux_node_type_id = ant.id
	AND an.ext_aux_node_id = aux_id
    AND an.cache_id = c_id;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_aux_type(bigint, bigint, refcursor)
    OWNER TO i3dhub;





-- FUNCTION: l3d.get_aux_by_id(bigint, bigint, refcursor)

DROP FUNCTION IF EXISTS l3d.get_aux_by_id(bigint, bigint, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_by_id(
	c_id bigint,
	aux_id bigint,
	ref refcursor)
    RETURNS refcursor
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE

AS $BODY$


BEGIN
OPEN ref FOR SELECT avt.name as name, av.value_str as value, av.value_num as value
	FROM l3d.aux_value as av, l3d.aux_node as an, l3d.aux_value_name as avt, l3d.aux_node_type as ant
	WHERE av.aux_value_name_id = avt.id
	AND an.id = av.aux_node_id
	AND an.aux_node_type_id = ant.id
	AND an.id =  (SELECT id FROM l3d.aux_node as an WHERE an.ext_aux_node_id = aux_id AND an.cache_id = c_id)
	ORDER BY an.id;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_aux_by_id(bigint, bigint, refcursor)
    OWNER TO i3dhub;



-- FUNCTION: l3d.get_aux_attributes_by_id(bigint, bigint, refcursor)

DROP FUNCTION IF EXISTS l3d.get_aux_attributes_by_id(bigint, bigint, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_attributes_by_id(
	c_id bigint,
	aux_id bigint,
	ref refcursor)
    RETURNS refcursor
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE

AS $BODY$

BEGIN
OPEN ref FOR SELECT aat.key as name, aat.value as value
	FROM l3d.aux_attributes as aat, l3d.aux_node as an
	WHERE an.id = aat.aux_node_id
	AND an.id =  (SELECT id FROM l3d.aux_node as an WHERE an.ext_aux_node_id = aux_id AND an.cache_id = c_id)
	ORDER BY an.id;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_aux_attributes_by_id(bigint, bigint, refcursor)
    OWNER TO i3dhub;




-- FUNCTION: l3d.get_aux_successor_by_id(bigint, bigint, refcursor)

DROP FUNCTION IF EXISTS l3d.get_aux_successor_by_id(bigint, bigint, integer, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_successor_by_id(
	c_id bigint,
	aux_id bigint,
	off_set integer,
	ref refcursor)
    RETURNS refcursor
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE

AS $BODY$

BEGIN
OPEN ref FOR SELECT an.ext_aux_node_id  + off_set as "nodeId"
             FROM l3d.aux_node as an, l3d.aux_node_to_aux_node as ansp
             WHERE ansp.pred_id = (SELECT id FROM l3d.aux_node as an WHERE an.ext_aux_node_id = aux_id AND an.cache_id = c_id)
             AND ansp.succ_id = an.id
	         ORDER BY an.ext_aux_node_id;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_aux_successor_by_id(bigint, bigint, integer, refcursor)
    OWNER TO i3dhub;




-- FUNCTION: l3d.get_aux_predecessor_by_id(bigint, bigint, refcursor)

DROP FUNCTION IF EXISTS l3d.get_aux_predecessor_by_id(bigint, bigint, integer, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_predecessor_by_id(
	c_id bigint,
	aux_id bigint,
	off_set integer,
	ref refcursor)
    RETURNS refcursor
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE

AS $BODY$

BEGIN
OPEN ref FOR SELECT an.ext_aux_node_id + off_set as "nodeId" FROM l3d.aux_node as an, l3d.aux_node_to_aux_node as ansp
             WHERE ansp.succ_id = (SELECT id FROM l3d.aux_node as an WHERE ext_aux_node_id = aux_id AND an.cache_id = c_id)
             AND ansp.pred_id = an.id
             ORDER BY an.ext_aux_node_id;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_aux_predecessor_by_id(bigint, bigint,integer, refcursor)
    OWNER TO i3dhub;



-- FUNCTION: l3d.get_aux_by_ids(bigint, bigint[], refcursor)

DROP FUNCTION IF EXISTS l3d.get_aux_by_ids(bigint, bigint[], refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_by_ids(
	c_id bigint,
	aux_ids bigint[],
	ref refcursor)
    RETURNS refcursor
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE

AS $BODY$


BEGIN
OPEN ref FOR SELECT an.ext_aux_node_id ,ant.name, avt.name , av.value
	FROM l3d.aux_value as av, l3d.aux_node as an, l3d.aux_value_name as avt, l3d.aux_node_type as ant
	WHERE av.aux_value_name_id = avt.id
	AND an.id = av.aux_node_id
	AND an.aux_node_type_id = ant.id
	AND an.ext_aux_node_id = ANY($2::bigint[])
    AND an.cache_id = c_id
	ORDER BY an.id;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_aux_by_ids(bigint, bigint[], refcursor)
    OWNER TO i3dhub;




-- FUNCTION: l3d.su_del_cache_by_ids(bigint[])

DROP FUNCTION IF EXISTS l3d.su_del_cache_by_ids(bigint[]);

CREATE OR REPLACE FUNCTION l3d.su_del_cache_by_ids(c_ids bigint[])
RETURNS void
LANGUAGE plpgsql
AS $function$
declare
          v_cache_ids integer[] := c_ids::integer[];

          v_node_ids bigint[];
          v_aux_node_ids bigint[];

          v_edge_instance_id bigint[];
          v_face_instance_id bigint[];
BEGIN

   select array_agg(id) into v_node_ids from l3d.node WHERE cache_id = ANY(v_cache_ids);
   select array_agg(id) into v_aux_node_ids from aux_node WHERE cache_id = ANY(v_cache_ids);

   select array_agg(id) into v_edge_instance_id from edge_instance WHERE cache_id = ANY(v_cache_ids);
   select array_agg(id) into v_face_instance_id from face_instance WHERE cache_id = ANY(v_cache_ids);


   DELETE FROM l3d.metadata WHERE node_id = any(v_node_ids);

   DELETE FROM l3d.tags_values WHERE node_id = any(v_node_ids);

   DELETE FROM l3d.aux_attributes where aux_node_id = any(v_aux_node_ids);

   DELETE FROM l3d.aux_value WHERE aux_node_id = any(v_aux_node_ids);

   delete from l3d.foreign_face_link ffl where ffl.aux_node_id = any(v_aux_node_ids);


   DELETE FROM l3d.edge_instance_to_aux_node WHERE aux_node_id = any(v_aux_node_ids) and edge_instance_id = any(v_edge_instance_id);
   DELETE FROM l3d.aux_node_to_aux_node WHERE succ_id = any(v_aux_node_ids) and pred_id  =  any(v_aux_node_ids);

   DELETE FROM l3d.edge_instance_to_aux_node WHERE aux_node_id = any(v_aux_node_ids) and edge_instance_id = any(v_edge_instance_id);
   DELETE FROM l3d.face_instance_to_aux_node WHERE aux_node_id = any(v_aux_node_ids) and face_instance_id = any(v_face_instance_id);

   delete from l3d.face_instance_to_original fio where fio.face_instance_id  = any(v_face_instance_id);

   delete from l3d.edge_instance where id  = any(v_edge_instance_id);

   delete from l3d.face_instance where id  = any(v_face_instance_id);

   DELETE FROM l3d.shape_instance WHERE cache_id = ANY(v_cache_ids);

   DELETE FROM l3d.edge WHERE cache_id = ANY(v_cache_ids);
   DELETE FROM l3d.face WHERE cache_id = ANY(v_cache_ids);

   DELETE FROM l3d.shape WHERE cache_id = ANY(v_cache_ids);

   DELETE FROM l3d.data_stats WHERE cache_id= ANY(v_cache_ids);

   DELETE FROM l3d.index_cache WHERE cache_id = ANY(v_cache_ids);
   DELETE FROM l3d.index_properties WHERE cache_id = ANY(v_cache_ids);
   DELETE FROM l3d.index_source WHERE cache_id = ANY(v_cache_ids);
   DELETE FROM l3d.index_stats WHERE cache_id = ANY(v_cache_ids);


   DELETE FROM l3d.node WHERE id = ANY(v_node_ids);
   DELETE FROM l3d.aux_node WHERE id  = any(v_aux_node_ids);

   DELETE FROM l3d.cache WHERE id = ANY(v_cache_ids);

END;
$function$;


ALTER FUNCTION l3d.su_del_cache_by_ids(bigint[])
    OWNER TO i3dhub;



-- FUNCTION: l3d.deactivate_cache(bigint)

DROP FUNCTION IF EXISTS l3d.deactivate_cache(bigint);

CREATE OR REPLACE FUNCTION l3d.deactivate_cache(
	c_id bigint)
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$
BEGIN
UPDATE l3d.cache SET active=false
WHERE id =$1;
END;

$BODY$;

ALTER FUNCTION l3d.deactivate_cache(bigint)
    OWNER TO i3dhub;

-- FUNCTION: l3d.deactivate_caches()

DROP FUNCTION IF EXISTS l3d.deactivate_caches(text[]);

CREATE OR REPLACE FUNCTION l3d.deactivate_caches(p_uuids text[])
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$
BEGIN
UPDATE l3d.cache SET active=false
WHERE uuid = any(p_uuids);
END;

$BODY$;

ALTER FUNCTION l3d.deactivate_caches(text[])
    OWNER TO i3dhub;


DROP FUNCTION IF EXISTS l3d.delete_provider_data(character varying);

CREATE OR REPLACE FUNCTION l3d.delete_provider_data(
	p_uuid character varying)
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$
BEGIN
	--deactivate caches
	UPDATE l3d.cache SET active=false
	WHERE uuid =$1;
	--delete indexl3d
	DELETE FROM l3d.index_cache
	WHERE cache_id IN (
		SELECT id FROM l3d.cache
	 	WHERE uuid = p_uuid
	);

	DELETE FROM l3d.index_properties
	WHERE cache_id IN (
		SELECT id FROM l3d.cache
	 	WHERE uuid = p_uuid
	);


	DELETE FROM l3d.index_source
	WHERE cache_id IN (
		SELECT id FROM l3d.cache
	 	WHERE uuid = p_uuid
	);


	DELETE FROM l3d.index_stats
	WHERE cache_id IN (
		SELECT id FROM l3d.cache
	 	WHERE uuid = p_uuid
	);
END;

$BODY$;

ALTER FUNCTION l3d.delete_provider_data(character varying)
    OWNER TO i3dhub;




-- FUNCTION: l3d.get_aux_types(refcursor)

DROP FUNCTION IF EXISTS l3d.get_aux_types(refcursor);

CREATE OR REPLACE FUNCTION l3d.get_aux_types(
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$

BEGIN
OPEN ref FOR SELECT ant.name as name
FROM l3d.aux_node_type as ant;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_aux_types(refcursor)
    OWNER TO i3dhub;



-- FUNCTION: l3d.get_statsPerAux(character varying, character varying, refcursor)
DROP FUNCTION IF EXISTS l3d.get_statsPerAux(character varying,character varying, bigint, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_statsPerAux(
	auxType character varying,
    valueType character varying,
	cid bigint,
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$

BEGIN
OPEN ref FOR SELECT  value_str as name, COUNT (value_str) as count
	FROM l3d.aux_value as av, l3d.aux_node as an
    WHERE  av.aux_node_id= an.id
    AND an.aux_node_type_id = (SELECT id FROM l3d.aux_node_type WHERE name = auxType)
    AND av.aux_value_name_id = (SELECT id FROM l3d.aux_value_name WHERE name = valueType)
	AND an.cache_id = cid
    AND value_num is NULL
    GROUP BY value_str;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_statsPerAux(character varying, character varying, bigint, refcursor)
    OWNER TO i3dhub;








-- FUNCTION: l3d.get_statsperaux(character varying, character varying, bigint, refcursor)

DROP FUNCTION IF EXISTS  l3d.search_aux(auxtype character varying, valuetype character varying, compareValue character varying,	ref refcursor);

CREATE OR REPLACE FUNCTION l3d.search_aux(
	auxtype character varying,
	valuetype character varying,
    compareValue character varying,
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
	AS $BODY$

BEGIN
OPEN ref FOR SELECT  an.id,value_str as value
	FROM l3d.aux_value as av, l3d.aux_node as an
    WHERE  av.aux_node_id= an.id
    AND an.aux_node_type_id = (SELECT id FROM l3d.aux_node_type WHERE name = auxType)
    AND av.aux_value_name_id = (SELECT id FROM l3d.aux_value_name WHERE name = valueType)
    AND an.cache_id = cid
    AND av.value_str = compareValue;

RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.search_aux(character varying, character varying, character varying, refcursor)
    OWNER TO i3dhub;


-- FUNCTION: l3d.get_statsperaux(bigint, bigint, refcursor)

DROP FUNCTION IF EXISTS l3d.get_auxstatsperaux(bigint, bigint, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_auxstatsperaux(
	auxId bigint,
	cid bigint,
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE


AS $BODY$

BEGIN
OPEN ref FOR SELECT ant.name as name, COUNT ( asp.succ_id) as count
	FROM l3d.aux_node_to_aux_node as asp , l3d.aux_node as an, l3d.aux_node_type as ant

    WHERE asp.succ_id = an.id
    AND an.aux_node_type_id = ant.id
  	AND asp.pred_id = (SELECT id FROM l3d.aux_node as an WHERE aux_node_id = auxId AND an.cache_id = cid)
    GROUP BY ant.name;

RETURN ref;
END;

$BODY$;


ALTER FUNCTION l3d.get_auxstatsperaux(bigint, bigint, refcursor)
    OWNER TO i3dhub;


DROP FUNCTION IF EXISTS l3d.search_aux_label(bigint, character varying, character varying, character varying, integer, refcursor);

CREATE OR REPLACE FUNCTION l3d.search_aux_label(
	cid bigint,
	auxtype character varying,
	op character varying,
	labelstr character varying,
	off_set integer,
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$

BEGIN
IF op = 'contains' THEN
	OPEN ref FOR SELECT aux_node_id + off_set as "nodeId", "Label" as "label"
	FROM l3d.aux_node as an, l3d.aux_node_type as ant
	WHERE an.aux_node_type_id = ant.id
    AND ant.name = auxType
    AND an.cache_id = cid
	AND "Label" LIKE '%'||labelStr||'%';

	ELSE IF op = 'equals' THEN
    OPEN ref FOR SELECT aux_node_id + off_set as "nodeId", "Label" as "label"
	FROM l3d.aux_node as an, l3d.aux_node_type as ant
    WHERE an.aux_node_type_id = ant.id
    AND ant.name = auxType
    AND an.cache_id = cid
    AND "Label" ~*  replace('^'||labelStr,'*','.*');
    END IF;
END IF;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.search_aux_label(bigint, character varying, character varying, character varying, integer, refcursor)
    OWNER TO i3dhub;


-- FUNCTION: l3d.update_aux_temp_table(bigint)

DROP FUNCTION IF EXISTS l3d.update_aux_temp_table(bigint);

CREATE OR REPLACE FUNCTION l3d.update_aux_temp_table(
	auxid bigint)
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$

BEGIN

PERFORM * FROM temp_auxIDs WHERE temp_auxIDS.AUXNode_ID = auxid;
    IF FOUND THEN
       UPDATE temp_auxIDS SET matching =matching +1 WHERE temp_auxIDS.AUXNode_ID = auxid;

    ELSE
       INSERT INTO temp_auxIDS (AUXNode_ID , matching) VALUES (auxid, 1);
    END IF;

END;

$BODY$;

ALTER FUNCTION l3d.update_aux_temp_table(bigint)
    OWNER TO i3dhub;

-- FUNCTION: l3d.search_aux_nodes(character varying[], character varying[], character varying[], integer, bigint, integer[])

DROP FUNCTION IF EXISTS l3d.search_aux_nodes(character varying[], character varying[], character varying[], integer, bigint, integer[]);

drop function if exists l3d.search_aux_nodes(varchar[],varchar[],varchar[],integer[],bigint[]);

CREATE OR REPLACE FUNCTION l3d.search_aux_nodes(p_keys varchar[], p_relations varchar[], p_values varchar[], p_caches_ids integer[], p_nodes bigint[] DEFAULT ARRAY[]::bigint[])
 RETURNS TABLE(id bigint, nodeid integer, label varchar, auxtype varchar, cache_id integer)
 LANGUAGE plpgsql
AS $function$
declare

	v_properties_count	integer;
	v_attributes_count	integer;
	v_ext_node_id		integer;

	v_criteria_count	integer;

	v_nodes				integer[];
	v_nodes_count		integer;
begin

--		raise notice 'select * from l3d.search_aux_nodes(''%''::varchar[],''%''::varchar[],''%''::varchar[],''%''::integer[],''%''::bigint[])', p_keys, p_relations, p_values,  p_caches_ids, p_nodes;

	select coalesce(p_nodes, array[]::integer[]) into v_nodes;

	select coalesce(array_length(v_nodes, 1), 0) into v_nodes_count;


	select count(*) into v_properties_count from unnest(p_keys) as keys where keys not ilike '%pmiType' and keys ilike 'auxProperties.%';

	select count(*) into v_attributes_count from unnest(p_keys) as keys where keys ilike 'auxAttributes.%';

	select coalesce(min(cast(p_values[idx] as integer)), -1) into v_ext_node_id
	from
	(
		select row_number() over() as idx, elements as name from unnest(p_keys) as elements
	) as keys
	where keys.name ilike 'nodeId';

	v_criteria_count := v_attributes_count + v_properties_count;

--	raise notice 'v_pmi_type: %, v_criteria_count: % v_properties_count % v_attributes_count % p_caches_ids % p_keys %', v_pmi_type, v_criteria_count, v_properties_count, v_attributes_count , p_caches_ids, p_keys;

	return query
	with sel_node_type as
	(
		select ant.id , ant.name
		from 
			l3d.aux_node_type ant
		where 
		exists
		(
			select count(*)	
			from
			(
				select o, ( '{'	|| v || '}')::varchar[] as v
				from
				(		
					select o, case when v ilike '[%]' then  substring(v , '[[](.*)[]]')  when o ilike '%equals' then ( v || '%') when o ilike '%contains' then ('%' || v || '%')  end as v
					from
					(
						select k.v as k, o.v as o, replace( v.v , '*', '%') as v		
						from
								( select row_number() over() as i, k as v from unnest(p_keys) as k ) k,
								( select row_number() over() as i, o as v from unnest(p_relations) as o ) o,		
								( select row_number() over() as i, o as v from unnest(p_values) as o ) v				
						where k.i = o.i and k.i = v.i and k.v ilike '%pmiType'		
					) a		
				) a
			) a 
			where  
				  case when o ilike 'not%' then ant."name" not ilike all(v) else ant."name" ilike any(v)  end		  
			having count(*)	 = (select count(*) from unnest(p_keys) k  where k ilike '%pmiType')	
		)
	)	
	select a.id, min(a.ext_aux_node_id)   as aux_node_id, cast(min(a.label) as varchar) as label, cast(min(a.name) as varchar) as name, min(a.cache_id)
	from
	(

		SELECT nod.id, nod.ext_aux_node_id, nod.label, typ.name,  1 as matches, nod.cache_id
		FROM
			aux_node as nod,
			sel_node_type typ
		where
			v_criteria_count = 0																		and
			( ( v_nodes_count = 0 and nod.cache_id = any(p_caches_ids) ) or nod.id = any(v_nodes) ) 	and
			(v_ext_node_id = -1 or nod.ext_aux_node_id = v_ext_node_id) 								and
			typ.id = nod.aux_node_type_id

		union all

		select nod.id, nod.ext_aux_node_id, nod.label, min(typ.name), count(*) as matches, nod.cache_id
		from
			l3d.aux_node as nod,
			sel_node_type as typ,
			l3d.aux_value_name as nam,
			l3d.aux_value as val,
			(
				select 	key,
					 	relation,
					 	case when value ~ '^(([1-9]\d*)|0)((\.|,)\d+)?$' and value not like '0x%' then replace(value, ',', '.')::numeric else null end as value_num,
					 	case when relation = 'contains' then ('%' || value || '%') else value end as value_str,
						case when relation in ('equalsAny', 'equals') then value::varchar[] else null end as array_str,
					 	case when relation in ('equalsAny', 'equals') then
					 			(
					 				select array_agg(replace(v, ',', '.')::numeric)
					 				from unnest( value::varchar[]) as v where v ~ '^(([1-9]\d*)|0)((\.|,)\d+)?$' and v not like '0x%'
					 			)
					 		else
					 			null
					 		end as array_num
				from
				(
					select key, relation,
							case when value like '[%]' then
								('{' || rtrim( ltrim(value, '['), ']') || '}')
								when relation = 'equals' and value not like '{%}' then
								('{' || value || '}')
								else
								value
							end as value
					from
					(
						select replace(key.name, 'auxProperties.', '') as key,
								relation.name as relation,
								replace(trim(value.name, '"'),'*','%') as value
						from
							(select row_number() over() as id, elements as name from unnest(p_keys) as elements) as key,
							(select row_number() over() as id, elements as name from unnest(p_relations) as elements) as relation,
							(select row_number() over() as id, trim(elements, '"') as name from unnest(p_values) as elements) as value
						where
							key.name not ilike '%pmiType' and key.name ilike 'auxProperties.%' and key.id = relation.id and  key.id = value.id
					) a

				) as a
			) as cnd
		where
			v_properties_count > 0																			and
			( ( v_nodes_count = 0 and nod.cache_id = any( p_caches_ids ) ) or nod.id = any( v_nodes ) ) 	and
			(v_ext_node_id = -1 or nod.ext_aux_node_id = v_ext_node_id) 									and
			
			nod.aux_node_type_id = typ.id 		and
			val.aux_node_id = nod.id 			and
			nam.id = val.aux_value_name_id 		and

			nam.name = cnd.key  				and
			(
				(
					cnd.relation in('equals', 'equalsAny') and
					(
						(
							cnd.array_num is not null and
							(
								(
									val.value_num is not null and val.value_num = any( cnd.array_num )
								) or
								(
									-- this is necessary, since select '0,4' < '0.4' => true , therefore a numeric conversion is required
									val.value_str is not null and val.value_str ~ '^(([1-9]\d*)|0)((\.|,)\d+)?$'  and replace(val.value_str, ',', '.')::numeric = any( cnd.array_num )
								)
							)
						) or
						(
							val.value_str is not null and  cnd.array_num is null and
							exists
							(
								select 1 from unnest(cnd.array_str) v where val.value_str ilike v or val.value_str ~*  ( '^((\\")|"|'')?' || replace(v, '%', '.*') || '((\\")|"|'')?$')
							)
						)
					)
				) or
				(
					cnd.relation = 'contains' and
					(
						(val.value_str is not null and val.value_str ilike cnd.value_str) or
						(val.value_num is not null and val.value_num::varchar ilike cnd.value_str)
					)
				) or
				(
					cnd.value_num is not null and
					(
						(
							val.value_num is not null and
							(
								(cnd.relation in ('lessOrEqualThan', 'greaterOrEqualThan')	and val.value_num = cnd.value_num) or
								(cnd.relation in ('lessThan', 'lessOrEqualThan') 						and val.value_num < cnd.value_num) or
								(cnd.relation in ('greaterThan', 'greaterOrEqualThan')  				and val.value_num > cnd.value_num)
							)
						) or
						(
							-- this is necessary, since select '0,4' < '0.4' => true , therefore a numeric conversion is required
							val.value_str is not null and val.value_str ~ '^(([1-9]\d*)|0)((\.|,)\d+)?$' and val.value_num is null and
							(
								(cnd.relation in ('lessOrEqualThan', 'greaterOrEqualThan')	and replace(val.value_str, ',', '.')::numeric = cnd.value_num)		or
								(cnd.relation in ('lessThan', 'lessOrEqualThan') 						and replace(val.value_str, ',', '.')::numeric < cnd.value_num)		or
								(cnd.relation in ('greaterThan', 'greaterOrEqualThan')  				and replace(val.value_str, ',', '.')::numeric > cnd.value_num)
							)
						)
					)
				)
				or
				(
					val.value_str is not null and  cnd.value_num is null and
					(
						(cnd.relation in ('lessOrEqualThan', 'greaterOrEqualThan')	and val.value_str = cnd.value_str)					or
						(cnd.relation in ('lessThan', 'lessOrEqualThan') 						and val.value_str < cnd.value_str)		or
						(cnd.relation in ('greaterThan', 'greaterOrEqualThan')  				and val.value_str > cnd.value_str) 		or
						(cnd.relation in ('equalss')											and val.value_str ~  ( '^((\\")|"|'')?' || replace(cnd.value_str, '%', '.*') || '((\\")|"|'')?$') )
					)
				)
			)
		group by nod.id
		having count(*) = v_properties_count

		union all

		SELECT nod.id, nod.ext_aux_node_id, nod.label, min(typ.name) as name, count(*) as matches, nod.cache_id
			FROM
				aux_node as nod,
				sel_node_type as typ,
				aux_attributes as att,
				(
					select key, relation , value , case when relation in('equals', 'equalsAny') then value::varchar[] else null end as value_array
					from
					(
						select  key, relation ,
							case
								when value like '[%]' then ('{' || rtrim( ltrim(value, '['), ']') || '}')
								 when relation = 'contains' then ('%' || value || '%')
								 when relation = 'equals'  then ('{' || value || '}')
								 else value
								end as value
						from
						(
							select replace(key.elements, 'auxAttributes.', '') as key,
								  relation.elements as relation ,

								  replace(trim(value.elements, '"'), '*', '%') as value
							from
								(select row_number() over() as id, * from unnest(p_keys) as elements) as key,
								(select row_number() over() as id, * from unnest(p_relations) as elements) as relation,
								(select row_number() over() as id, * from unnest(p_values) as elements) as value
							where
							key.elements ilike 'auxAttributes.%' and key.id = relation.id and key.id = value.id
						) as a

					) a
				) as cnd
			WHERE
				v_attributes_count > 0 																		and
				( ( v_nodes_count = 0 and nod.cache_id = any( p_caches_ids ) ) or nod.id = any(v_nodes) ) 	and
				(v_ext_node_id = -1 or nod.ext_aux_node_id = v_ext_node_id) 								and				

				nod.aux_node_type_id = typ.id 		and
				att.aux_node_id = nod.id 			and

				att.key = cnd.key  					and
				(
					(cnd.relation = 'contains'				and att.value ilike cnd.value)					or
					(cnd.relation = 'equalss'				and att.value ~ ( '^((\\")|"|'')?' || replace(cnd.value, '%', '.*') || '((\\")|"|'')?$')  )		or
					(cnd.relation = 'lessThan' 				and att.value < cnd.value) 						or
					(cnd.relation = 'lessOrEqualThan' 		and	att.value <= cnd.value)						or
					(cnd.relation = 'greaterThan' 			and att.value > cnd.value)						or
					(cnd.relation = 'greaterOrEqualThan'	and att.value >= cnd.value)						or
					(cnd.relation in('equals', 'equalsAny')	and exists(select * from unnest(cnd.value_array) as v where att.value ~* ( '^((\\")|"|'')?' || replace(v, '%', '.*') || '((\\")|"|'')?$' ) ))
				)
			group by nod.id
			having count(*) = v_attributes_count

	) as a
	group by a.id
	having sum(matches) >=  v_criteria_count;
end;
$function$;

ALTER FUNCTION l3d.search_aux_nodes(varchar[],varchar[],varchar[],integer[],bigint[]) OWNER TO i3dhub;


drop function if exists l3d.search_query_aux(varchar[], varchar[], varchar[], varchar[], varchar[], integer[], bigint[]);
drop function if exists l3d.search_query_aux(varchar[], varchar[], varchar[], varchar[], varchar[], integer[]);

create or replace function l3d.search_query_aux(p_fields varchar[], p_keys varchar[], p_operators varchar[], p_values varchar[], p_offsets varchar[], p_caches_ids integer[])
returns table(query_result varchar[])
    LANGUAGE 'plpgsql'
AS $BODY$
declare
    v_fields varchar[];
    v_properties varchar[];
    v_attributes varchar[];

    v_nodes bigint[] := array[]::bigint[];

    v_offsets varchar[] := p_offsets;
    v_shape_instance_offset integer := 0;

	v_search_face_link boolean;   
	v_search_face boolean;
    v_search_edge boolean;   
   
	v_node_types varchar[]; 
	v_exclude_node_types boolean;
   

begin
	
	v_search_face_link := exists( select * from unnest(p_fields) k where k like '%FaceLink%' ) ;
	
-- raise notice 'select * from l3d.search_query_aux(''%'',''%''''%''''%'',%,%)', p_fields, p_keys, p_operators, p_values, p_offsets, p_caches_ids;
        -- offsets
    if array_length( p_caches_ids, 1 ) = 1 then
        if v_offsets[1] is not null and v_offsets[1] like '%shapeInstanceOffset=>%' then
            v_shape_instance_offset := coalesce( coalesce( v_offsets[1]::hstore, 'shapeInstanceOffset=>0'::hstore )->'shapeInstanceOffset', '0' ) ::integer;
        end if;

        if v_offsets[1] is null or v_offsets[1] not like '%auxOffset=>%' then
            v_offsets[1] := '0,0';
        else
            declare
                v_offset integer := coalesce( coalesce( v_offsets[1]::hstore, 'auxOffset=>0'::hstore )->'auxOffset', '0' ) ::integer;
            begin
                v_offsets[1] := v_offset::varchar || ',' || v_offset::varchar;
            end;
        end if;
    end if;
   
	select o ilike 'not%' ,( '{'	|| v || '}')::varchar[] into
	v_exclude_node_types, v_node_types
	from
	(		
		select o, case when v ilike '[%]' then  substring(v , '[[](.*)[]]')  when o ilike '%equals' then ( v || '%') when o ilike '%contains' then ('%' || v || '%')  end as v
		from
		(
			select k.v as k, o.v as o, replace( v.v , '*', '%') as v		
			from
					( select row_number() over() as i, k as v from unnest(p_keys) as k ) k,
					( select row_number() over() as i, o as v from unnest(p_operators) as o ) o,		
					( select row_number() over() as i, o as v from unnest(p_values) as o ) v				
			where k.i = o.i and k.i = v.i and k.v ilike '%pmiType'		
		) a		
	) a ;
   
-- raise warning 'v_node_types % v_exclude_node_types %', v_node_types, v_exclude_node_types;


    if exists(select * from unnest(p_keys) k where k ilike '%FaceLink') then
        declare
            v_face_links varchar[];
            v_face_link varchar;
        begin
            select replace( p_values[i] , '\', '%') into v_face_link from ( select row_number() over() as i, v from unnest(p_keys) v ) k  where k.v ilike '%FaceLink';

            v_face_links := case when v_face_link like '[%]' then string_to_array( left( right( v_face_link , -1 ), -1 ) ,  ',' ) else array[v_face_link] end ;

            select array_agg( a.id ) into v_nodes
            from
            (
                select an.id
                from l3d.foreign_face_link ffl , l3d.aux_node an
                where
                    ffl.aux_node_id = an.id                         and
                    an.cache_id = any( p_caches_ids )               and
                    ffl.face_link like any( v_face_links )
            ) a;
        end;
    elsif 'pointsTo' ilike any(p_keys) then

        declare
                v_ext_node_id integer;
        begin

            select ( p_values[i] )::integer into v_ext_node_id from ( select row_number() over() as i, v from unnest(p_keys) v ) k  where k.v ilike 'pointsTo';

            select array_agg( a.node_id ) into v_nodes
            from
            (
                select pd_n.id as node_id
                from
                    l3d.aux_node pt_n,
                    l3d.aux_node_to_aux_node as ntn,
                    l3d.aux_node as pd_n
                WHERE
                    pt_n.cache_id = any( p_caches_ids ) and
                    pt_n.ext_aux_node_id = v_ext_node_id and

                    ntn.succ_id = pt_n.id and
                    pd_n.id = ntn.pred_id

                order by pd_n.ext_aux_node_id
            ) a;
        end;
    elsif 'faceHandle' ilike any(p_keys) then

        declare
            v_face_id integer;
            v_shape_instance_id integer;
            v_faceHandle hstore;
        begin

            select (replace(p_values[i], ':', '=>') ) ::hstore into v_faceHandle from ( select row_number() over() as i, v from unnest(p_keys) v) k where k.v ilike 'faceHandle';

            select value into v_shape_instance_id from each(v_faceHandle) where key ilike '%shapeInstanceId%';

            select value into v_face_id from each(v_faceHandle) where key ilike '%faceId%';

            if v_face_id is null then
                select value into v_face_id from each(v_faceHandle) where key ilike '%entityId%';
            end if;

            if v_face_id is null then
                raise exception 'Missing entityId/FaceId in faceHandle';
            end if;
/*
            select array_agg(fia.aux_node_id) into v_nodes
            from
                l3d.face fac, l3d.shape_instance shi, l3d.face_instance fin, l3d.face_instance_to_aux_node fia
            where
                fac.cache_id = any( p_caches_ids )              and
                fac.ext_face_id = v_face_id                     and

                shi.cache_id = any( p_caches_ids )              and
                shi.shape_instance_id = v_shape_instance_id     and

                fin.cache_id = any( p_caches_ids )              and
                fin.shape_instance_id = shi.id                  and
                fin.face_id = fac.id                            and

                fia.face_instance_id = fin.id;
*/
           
            select array_agg(fia.aux_node_id) into v_nodes
            from
                l3d.face fac, l3d.shape_instance shi, l3d.face_instance fin, l3d.face_instance_to_aux_node fia, l3d.aux_node an, l3d.aux_node_type ant
            where
                fac.cache_id = any( p_caches_ids )              and
                fac.ext_face_id = v_face_id                     and

                shi.cache_id = any( p_caches_ids )              and
                shi.shape_instance_id = v_shape_instance_id     and

                fin.cache_id = any( p_caches_ids )              and
                fin.shape_instance_id = shi.id                  and
                fin.face_id = fac.id                            and

                fia.face_instance_id = fin.id					and
                
                an.id = fia.aux_node_id							and
                an.aux_node_type_id = ant.id 					and
				( 	v_node_types is null or 
					(														
						( v_exclude_node_types and ant.name not ilike all(v_node_types) ) or ( not v_exclude_node_types and ant.name ilike any(v_node_types) ) 							
					)
				);

--             raise warning 'nodes %', v_nodes;

        end;
    end if;

    if v_nodes is null then
        return query select array[]::varchar[] from unnest(array[]::varchar[]);
    else
    
		v_search_face := exists ( select * from unnest(p_fields) k where k = any( array['faceHandle', 'faceId'] ) );
		v_search_edge := exists ( select * from unnest(p_fields) k where k = any( array['edgeHandle', 'edgeId'] ) );
    

        select coalesce( array_agg( replace(f, '*', '%') ), array[]::varchar[] )
        into v_properties
        from
        (
                select substring( f, char_length('metadata.auxProperties.') + 1 ) f from unnest( p_fields ) f where f ilike 'metadata.auxProperties.%'
        ) a;


        select coalesce( array_agg( replace(f, '*', '%') ) , array[]::varchar[] )
        into v_attributes
        from
        (
                select substring( f, char_length('metadata.auxAttributes.') + 1 ) f from unnest( p_fields ) f where f ilike 'metadata.auxAttributes.%'
        ) a;

        return query  select
        (
            select
                (
                    select
                    (
                        select a.fields_values
                        from
                        (
                            select
                            array_agg
                            (
                                case when f ilike 'nodeId'  then n.node_id::varchar
                                     when f ilike 'label'  then 
						(
							select 
								case when a.l  !~ '[^[:alnum:].]+' then 
										( '"' || a.l || '"' ) 
									else 
										right( left( array_to_json( array[a.l] )::varchar , -1 ), -1 ) 
								end
							from ( select trim( nod."label", '"')::varchar l ) a
						)
                                     
                                     when f ilike '%pmiType'  then ( '"' || nod.auxtype     || '"')
                                     when f ilike 'metadata.auxProperties.*' then '{' || array_to_string(prop.aux_values, ',') || '}'
                                     when f ilike 'metadata.auxProperties.%'  then
                                            coalesce(
                                            (
                                                select substring( mdf, position( ':' in mdf ) + 1 ) as v from unnest(prop.aux_values) mdf where mdf ilike ( '"' || f || '":%')
                                            ), '"undefined"')
                                    when f ilike 'metadata.auxAttributes.*' then '{' || array_to_string(att.aux_values, ',') || '}'
                                    when f ilike 'metadata.auxAttributes.%'  then
                                            coalesce(
                                            (
                                                select substring( mdf, position( ':' in mdf ) + 1 ) as v from unnest(att.aux_values) mdf where mdf ilike ( '"' || f || '":%')
                                            ), '"undefined"')
                                    when  f ilike '%FaceLink%' and ( ffl.face_link is not null or array_length(p_fields, 1) > 1 ) then
                                            ( '"' || replace( coalesce( ffl.face_link, 'undefined' ) , '"', '\"' ) || '"' )

                                    --------------------------------------------------------------------
                                    -- !!! IMPORTANT this check must come last, it will evaluate  !!! --
                                    -- !!! to true for every field and will cause the cases after !!! --
                                    -- !!! being passed over                                      !!! --
                                    --------------------------------------------------------------------
                                    when ( v_search_face or v_search_edge )  and ( fitan.face_instance_id is not null or eitan.edge_instance_id is not null or array_length( p_fields, 1 ) > 1 ) then

                                        case when fitan.face_instance_id is null and eitan.edge_instance_id is null and array_length( p_fields, 1 ) > 1 then
                                                '"undefined"'
                                        else
                                            (
                                                select
                                                    case when f ilike 'faceId' then
                                                        fc.ext_face_id::text
                                                    else
                                                        row_to_json( ( select r from ( select 1 as "entityType", si.shape_instance_id + v_shape_instance_offset as "shapeInstanceID", fc.ext_face_id as "entityID") r ) )::text
                                                    end
                                                from l3d.face_instance fi , l3d.shape_instance si, l3d.face fc
                                                where fi.id = fitan.face_instance_id
                                                and si.id = fi.shape_instance_id
                                                and fc.id = fi.face_id
                                                and f ilike any( array['faceHandle', 'faceId', 'topoHandle'] )

                                                union all

                                                select
                                                    case when f ilike 'edgeId' then
                                                        ed.ext_edge_id::text
                                                    else
                                                        row_to_json( ( select r from ( select 2 as "entityType", si.shape_instance_id + v_shape_instance_offset as "shapeInstanceID", ed.ext_edge_id as "entityID") r ) )::text
                                                    end
                                                from l3d.edge_instance ei, l3d.shape_instance si, l3d.edge ed
                                                where ei.id = eitan.edge_instance_id
                                                and si.id = ei.shape_instance_id
                                                and ed.id = ei.edge_id
                                                and f ilike any( array['edgeHandle', 'edgeId', 'topoHandle'] )
                                            )
                                        end
                                    end
                            ) as fields_values
                            from unnest(p_fields) f
                        ) a
                        where exists(select * from unnest(a.fields_values) fv where fv <> '"undefined"')
                    )
                    from
                    (
                        -- offsets : [29525,13704],[29549,13705],[29573,13706],[ ...
                        select nod.nodeid + split_part(o.offset_root_id, ',' , case when nod.nodeid != 0 then 1 else 2 end)::integer as node_id
                        from
                        (
                            select unnest( string_to_array( offsets, '],[') ) offset_root_id
                            from
                            (
                                select rtrim( ltrim(v_offsets[idx], '['), ']') offsets
                                from
                                    (select row_number() over() as idx, cache_id from unnest( p_caches_ids ) as cache_id) c
                                where
                                    c.cache_id = nod.cache_id
                            ) a
                        ) o
                    ) n
                )
            from
            (
                select array_agg
                (
                    '"' || f  || '":' ||
                    coalesce( a.vn, case when a.vs !~ '[^[:alnum:].]+' then  ( '"' || a.vs || '"' ) else right( left( array_to_json( array[a.vs] )::varchar , -1 ), -1 ) end )
                )  aux_values
                from
                (
                    select f, vn, vs from( select 'metadata.auxProperties.pmiType'::varchar as f, null::varchar as vn, nod.auxtype as vs ) a where 'pmiType' ilike any( v_properties )
                    union all
                    select  'metadata.auxProperties.' || avn.name as f,
                             av.value_num::varchar as vn,
                             case when av.value_num  is null then
                                    case when av.value_str like '/"%/"'  then trim(av.value_str, '\"')
                                            when av.value_str like '"%"'  then trim(av.value_str, '"')
                                            else av.value_str
                                    end
                                else null
                             end as vs
                    from l3d.aux_value av , l3d.aux_value_name avn
                    where av.aux_value_name_id = avn.id  and av.aux_node_id = nod.id and avn."name" like any( v_properties )  and array_length( v_properties, 1 ) > 0 	

                ) a
            ) prop,
            (
                select array_agg
                (
                    '"' || f  || '":' ||
                    coalesce( a.vn, case when a.vs !~ '[^[:alnum:].]+' then  ( '"' || a.vs || '"' ) else right( left( array_to_json( array[a.vs] )::varchar , -1 ), -1 ) end )
                )  aux_values
                from
                (
                    select 'metadata.auxAttributes.' || aa."key" as f,
                            null::varchar as vn,
                            case when aa.value like '/"%/"'  then trim(aa.value, '\"')
                                    when aa.value like '"%"'  then trim(aa.value, '"')
                                    else aa.value
                            end as vs
                    from l3d.aux_attributes aa
                            where
                            aa.aux_node_id = nod.id and aa."key" like any(v_attributes) and array_length( v_attributes, 1 ) > 0
                 ) a
            ) att
        )
        from l3d.search_aux_nodes(p_keys, p_operators, p_values, p_caches_ids, v_nodes) nod
        left join l3d.foreign_face_link ffl on ffl.aux_node_id = nod.id and v_search_face_link
        left join l3d.face_instance_to_aux_node fitan on fitan.aux_node_id = nod.id and v_search_face
        left join l3d.edge_instance_to_aux_node eitan on eitan.aux_node_id = nod.id and v_search_edge 
        ;

    end if;
end;
$BODY$;

ALTER FUNCTION l3d.search_query_aux(varchar[], varchar[], varchar[], varchar[], varchar[], integer[]) OWNER TO i3dhub;



drop function if exists l3d.search_query_structure(varchar[], varchar[], varchar[], varchar[], integer[], varchar[]);

create or replace function l3d.search_query_structure(p_fields varchar[], p_keys varchar[], p_operators varchar[], p_values varchar[], p_caches_ids integer[], p_offsets varchar[])
returns table(query_result varchar[])
    LANGUAGE 'plpgsql'
AS $BODY$
declare
    v_offsets hstore;

    v_ext_node_id integer;
    v_labels varchar[];

   	v_metadata_count integer;

   	v_fields varchar[];
	v_metadata_fields varchar[];
begin

--	raise notice 'search_query: select * from l3d.search_query_structure(''%'',''%'',''%'',''%'',''%'',''%'')',p_fields,p_keys, p_operators, p_values, p_caches_ids, p_offsets;

	select coalesce( min(p_values[i]::integer), -1 ) into v_ext_node_id from (select row_number() over() as i, k from unnest(p_keys) as k) a where k ilike 'nodeId';


	select (
			case 
				when v like '{%}' then v 
				else 
					('{' || case when v ~ '"?[[].*[]]"?' then  substring(v , '[[](.*)[]]') when o = 'contains' then  ('%' || v || '%') else v end || '}') 
			end
		)::varchar[]
	into v_labels
	from
	(
		select coalesce(min(replace(p_values[i], '*', '%')), '{%}') v, coalesce(min(p_operators[i]), 'contains') o
		from
			(select row_number() over() as i, k from unnest(p_keys) as k) a
		 where k ilike 'label'
	) a;

--	raise notice 'v_labels %', v_labels;

	select count(*) into v_metadata_count from unnest(p_keys) k  where k not ilike all( array['nodeId', 'label', 'nodeType'] );


	select array_agg( replace( regexp_replace( regexp_replace(f , 'metadata.', '', 'i' ), 'structMetadata.', '' , 'i') , '*' , '%') )
	into v_metadata_fields
	from unnest(p_fields) f
	where f ilike '%metadata.%' ;

	RETURN QUERY
	select
	(
		select
		(
			select array_agg
				(
					case when f ilike 'nodeId'  then n.node_id::varchar
						 when f ilike 'label'  then
							(
							 	select 
									case when a.l  !~ '[^[:alnum:].]+' then 
											( '"' || a.l || '"' ) 
										else 
											right( left( array_to_json( array[a.l] )::varchar , -1 ), -1 ) 
									end
							 	from ( select trim( t."label", '"')::varchar l ) a
						 	) 						 						 
						when f ilike 'metadata.structMetaData.*' then '{' || array_to_string(t.md_store, ',') || '}'
						else
							coalesce(
							(
								select substring( mdf, position( ':' in mdf ) + 1 ) as v from unnest(t.md_store) mdf where mdf ilike ( '"' || f || '":%')
							), '"undefined"')
						end
				)::varchar[]
			from unnest(p_fields) f
		)
		from
		(
			-- offsets : [29525,13704],[29549,13705],[29573,13706],[ ...
			select t."nodeId" + split_part(o.offset_root_id, ',' , case when t."nodeId" != 0 then 1 else 2 end)::integer as node_id
			from
			(
				select unnest( string_to_array( offsets, '],[') ) offset_root_id
				from
				(
					select rtrim( ltrim(p_offsets[idx], '['), ']') offsets
					from
						(select row_number() over() as idx, cache_id from unnest( p_caches_ids ) as cache_id) c
					where
						c.cache_id = t.cache_id
				) a
			) o
		) n
	)
	from
	(
		SELECT  t.id,
				t."nodeId" as "nodeId",
				t.label ,
				t.md_store,
				t.cache_id
		FROM
		(
			select d.id, d."nodeId", d.label, d.cache_id,
				(

					select array_agg
							(
								'"' || f  || '":' ||
								coalesce(a.vn, case when a.vs !~ '[^[:alnum:].]+' then  ( '"' || a.vs || '"') else right( left( array_to_json( array[a.vs] )::varchar , -1 ), -1 ) end)
							 )
					from
					(
						select 	'metadata.structMetaData.' || mda."key" as f,
								case when mda.value_num is not null then mda.value_num::varchar else null end as vn,
								case when mda.value_num is null then mda.value_str else null end as vs
							from l3d.metadata mda
							where
								mda.node_id = d.id and mda."key" ilike any (v_metadata_fields)
					) a
				) as md_store
			from
			(
				select nod.id, nod.ext_node_id as "nodeId", nod.cache_id, nod.label
				from node nod
				where
					v_metadata_count = 0 																				and
					( v_ext_node_id = -1  or nod.ext_node_id = v_ext_node_id)											and
					nod.label ilike any( v_labels )																		and
					nod.cache_id = any( p_caches_ids )

				union all

				(
					select nod.id, nod.ext_node_id as "nodeId", nod.cache_id, nod.label
					from
						node nod,
						l3d.metadata mda,
						(
							select 	replace( key, 'structMetaData.', '') as key,
								 	relation,
									case when value ~ '^(([1-9]\d*)|0)((\.|,)\d+)?$' then replace(value, ',', '.')::numeric else null end as value_num,
								 	case when relation = 'contains' then ('%' || replace(value, '/', '\') || '%') else value end as value_str,
									case when value like '[%]' then
											( '{' || rtrim( ltrim( value, '[' ), ']' ) || '}' )::varchar[]
										 when relation = 'equalsAny' then
											value::varchar[]
										else
											null
										end as value_array
							from
							(
								select replace( key.name, 'metadata.', '' ) as key, relation.name as relation, replace( trim( value.name, '"' ),'*','%' ) as value
								from
									(select row_number() over() as id, elements as name from unnest(p_keys) as elements) as key,
									(select row_number() over() as id, elements as name from unnest(p_operators) as elements) as relation,
									(select row_number() over() as id, elements as name from unnest(p_values) as elements) as value
								where
									key.name not ilike all( array['nodeId', 'label', 'nodeType'] ) and key.id = relation.id and  key.id = value.id
							) as a
						) as cnd
					where
						v_metadata_count <> 0 																				and
						( v_ext_node_id = -1  or nod.ext_node_id = v_ext_node_id)											and
						nod.label ilike any( v_labels )																		and
						nod.cache_id =  any( p_caches_ids) 																	and
						mda.node_id  = nod.id 																				and
						mda."key" = cnd.key																					and
						(
							( cnd.relation = 'equals' and ( mda.value_num = cnd.value_num or mda.value_str ilike cnd.value_str or ( mda.value_str ~* ( '^((\\")|"|'')' || replace(cnd.value_str, '%', '.*') || '((\\")|"|'')$' ) )  ) ) or
							( cnd.relation = 'equalss' and ( mda.value_num = cnd.value_num or mda.value_str like cnd.value_str or ( mda.value_str ~ ( '^((\\")|"|'')' || replace(cnd.value_str, '%', '.*') || '((\\")|"|'')$' ) )  ) ) or
							(
								cnd.relation = 'equalsAny' and exists(
												select v from unnest(cnd.value_array) as v
												where
													(mda.value_num is not null and v = mda.value_num::text) or
													mda.value_str ilike v  or
													mda.value_str ~* ('^((\\")|"|'')' || replace(v, '%', '.*') || '((\\")|"|'')$')
											)
							) or
							(
								cnd.relation = 'contains' and coalesce(mda.value_str, mda.value_num::varchar) ilike cnd.value_str
							) or
							(
								cnd.value_num is not null and mda.value_num is not null and
								(
									(cnd.relation in ('lessOrEqualThan', 'greaterOrEqualThan')	and mda.value_num = cnd.value_num) 				or
									(cnd.relation in ('lessThan', 'lessOrEqualThan') 						and mda.value_num < cnd.value_num) or
									(cnd.relation in ('greaterThan', 'greaterOrEqualThan')  				and mda.value_num > cnd.value_num)
								)
							) or
							(
								mda.value_str is not null and  mda.value_num is null and
								(
									(cnd.relation in ('lessOrEqualThan', 'greaterOrEqualThan')	and mda.value_str = cnd.value_str)					or
									(cnd.relation in ('lessThan', 'lessOrEqualThan') 						and mda.value_str < cnd.value_str)		or
									(cnd.relation in ('greaterThan', 'greaterOrEqualThan')  				and mda.value_str > cnd.value_str)
								)
							)
						)
						group by nod.id
						having count(*) = v_metadata_count
					)
				) d
			) t
	) t;
end;
$BODY$;

ALTER FUNCTION l3d.search_query_structure(varchar[], varchar[], varchar[], varchar[], integer[], varchar[]) OWNER TO i3dhub;


drop function if exists l3d.search_query_original_face(varchar[],varchar[],varchar[],varchar[],varchar[],integer[]);

create or replace function l3d.search_query_original_face(p_fields varchar[], p_keys varchar[], p_operators varchar[], p_values varchar[], p_offsets varchar[], p_caches_ids integer[])
returns table(query_result varchar[])
    LANGUAGE 'plpgsql'
AS $BODY$
declare
    v_shape_instance_offset integer := 0;
begin
                if 'originalFaceID' ilike any(p_keys) and 'faceHandle' ilike any(p_fields) then

                        declare
                                v_original_face_id integer;
                        begin

                                select p_values[i] into  v_original_face_id from ( select row_number() over() as i, v from unnest(p_keys) v ) k  where k.v ilike 'originalFaceID';

                                if array_length( p_caches_ids, 1 ) = 1 then
                                        if p_offsets[1] is not null and p_offsets[1] like '%shapeInstanceOffset=>%' then
                                                v_shape_instance_offset := coalesce( coalesce( p_offsets[1]::hstore, 'shapeInstanceOffset=>0'::hstore )->'shapeInstanceOffset', '0' ) ::integer;
                                        end if;
                                end if;

                                return query
                                select array[row_to_json( ( select r from (select 1 as "entityType", si.shape_instance_id + v_shape_instance_offset as "shapeInstanceID",  f.ext_face_id  as "entityID") r ) )::varchar]
                                from  l3d.face_instance fi , l3d.face_instance_to_original fito , l3d.shape_instance si , l3d.face f
                                where
                                        fi.cache_id = any( p_caches_ids ) and
                                        fito.face_instance_id  = fi.id  and
                                        fito.original_face_id  = v_original_face_id and
                                        fi.shape_instance_id = si.id and
                                        f.id = fi.face_id ;
                        end;

                elsif 'originalFaceID' ilike any(p_fields) and 'faceHandle' ilike any(p_keys) then

                        declare
                                v_face_id integer;
                                v_shape_instance_id integer;
                                v_nodes bigint [];
                                v_faceHandle hstore;
                        begin

                                select (replace(p_values[i], ':', '=>') ) ::hstore into v_faceHandle from ( select row_number() over() as i, v from unnest(p_keys) v) k where k.v ilike 'faceHandle';


                                select value::integer into v_shape_instance_id from each(v_faceHandle) where key ilike 'shapeInstanceID%';

                                select value into v_face_id from each(v_faceHandle) where key ilike any(array['faceID%', 'entityID%']);

                                if v_face_id is null then
                                        raise exception 'Missing entityID/faceId in faceHandle';
                                end if;

                                return query
                                select array[fito.original_face_id ::varchar]
                                from
                                        l3d.face fac, l3d.shape_instance shi, l3d.face_instance fin, l3d.face_instance_to_original fito
                                where
                                        fac.cache_id = any( p_caches_ids )                      and
                                        fac.ext_face_id = v_face_id                                     and

                                        shi.cache_id = any( p_caches_ids )                      and
                                        shi.shape_instance_id = v_shape_instance_id     and

                                        fin.cache_id = any( p_caches_ids )                      and
                                        fin.shape_instance_id = shi.id                          and
                                        fin.face_id = fac.id                                            and

                                        fito.face_instance_id = fin.id;

                        end;
        end if;
end;
$BODY$;

ALTER FUNCTION l3d.search_query_original_face(varchar[],varchar[],varchar[],varchar[],varchar[],integer[]) OWNER TO i3dhub;


drop function if exists l3d.search_query(varchar,varchar,boolean);

CREATE OR REPLACE FUNCTION l3d.search_query(p_uri character varying, p_query character varying, result_as_string boolean)
 RETURNS TABLE(query_result varchar)
 LANGUAGE plpgsql
AS $function$
declare
	--- select * from l3d.search_query('/5b2f17fa79e7cdc3-45d38cf7-4645-b994/0/151/data/search', '{"query":{"select":["nodeId","metadata.auxAttributes.label"],"conditions":[{"nodeType":"aux"},{"metadata":"auxProperties.pmiType","equals":"DatumFeatureSymbol"},{"metadata":"auxAttributes.label","equals":{"select":["metadata.auxProperties.primaryReference"],"conditions":[{"nodeType":"aux"},{"metadata":"auxProperties.pmiType","equals":"FeatureControlFrame"},{"nodeId":49}]}}]},"offsets":{"structOffset":1,"auxOffset":2,"shapeInstanceOffset":3}}')
		v_fields varchar[];
		v_keys varchar[] ;
		v_operators varchar[] ;
		v_values varchar[];
		v_offsets hstore;

		v_cache_id bigint;
begin

		if result_as_string then

			return query select ('[' || coalesce( array_to_string( array_agg(r.query_result), ','), '') || ']')::varchar from l3d.search_query(p_uri, p_query, false) r;

		else

			return query
			select ( '[' || array_to_string(r.query_result, ',') || ']' ) ::varchar as qr
			from l3d.search_query(p_uri, p_query) r;

	 	end if;

end;
$function$;

alter function l3d.search_query(varchar,varchar,boolean) owner to i3dhub;



drop function if exists l3d.search_query(varchar,varchar);

CREATE OR REPLACE FUNCTION l3d.search_query(p_uri varchar, p_query varchar)
RETURNS table(query_result varchar[])
    LANGUAGE 'plpgsql'
AS $BODY$
declare
	--- select * from l3d.search_query('/5b2f17fa79e7cdc3-45d38cf7-4645-b994/0/151/data/search', '{"query":{"select":["nodeId","metadata.auxAttributes.label"],"conditions":[{"nodeType":"aux"},{"metadata":"auxProperties.pmiType","equals":"DatumFeatureSymbol"},{"metadata":"auxAttributes.label","equals":{"select":["metadata.auxProperties.primaryReference"],"conditions":[{"nodeType":"aux"},{"metadata":"auxProperties.pmiType","equals":"FeatureControlFrame"},{"nodeId":49}]}}]},"offsets":{"structOffset":1,"auxOffset":2,"shapeInstanceOffset":3}}')
	-- select * from l3d.search_query('/5b2f17fa79e7cdc3-45d38cf7-4645-b994/0/151/data/search', '{"query":{"select":["nodeId","metadata.auxAttributes.label"],"conditions":[{"nodeType":"aux"},{"metadata":"auxProperties.pmiType","equals":"DatumFeatureSymbol"},{"metadata":"auxAttributes.label","equals":{"select":["metadata.auxProperties.primaryReference"],"conditions":[{"nodeType":"aux"},{"metadata":"auxProperties.pmiType","equals":"FeatureControlFrame"},{"nodeId":49}]}}]},"offsets":{"structOffset":1,"auxOffset":2,"shapeInstanceOffset":3}}'::text);
	-- select * from l3d.search_query('/5b2f17fa79e7cdc3-45d38cf7-4645-b994/0/21420/data/search','{"query":{"select":["nodeId"],"conditions":[{"nodeType":"aux"},{"metadata":"auxProperties.pmiType","equals":"ModelView"},{"pointsTo":18}]},"offsets":{"structOffset":0,"auxOffset":0,"shapeInstanceOffset":0}}')

		v_caches_ids integer[];
		v_offsets varchar[];

		v_node_type varchar;
begin
		-- -------------------------------------------------------------------------------------------------------------------------------------------------------------
		-- cache on the uri
		if p_uri ~ '/[a-fA-F0-9-]{35,52}/[0-9]+/[0-9]+/data/search' then

			select array_agg( cac.id )::integer[]
			into v_caches_ids
			from
			(
				select
						substring(p_uri , '/([a-fA-F0-9-]{35,52})/') as uuid,
						substring(p_uri , '/[a-fA-F0-9-]{35,52}/([0-9]+)/') as ext_cache_id0,
						substring(p_uri , '/[a-fA-F0-9-]{35,52}/[0-9]+/([0-9]+)') as ext_cache_id1
			) a , l3d."cache" cac
			where
				cac.uuid  = a.uuid and
				cac.ext_cache_id0 = a.ext_cache_id0 ::integer and
				cac.ext_cache_id1  = a.ext_cache_id1::integer and
				cac.active ;

--			raise notice 'offsets: %',  substring (p_query , '"?offsets"?[ ]*:[ ]*{([^}]*)');

			v_offsets := array[ replace( substring (p_query , '"?offsets"?[ ]*:[ ]*{([^}]*)'), ':', '=>' )]::varchar[];


		elsif p_uri ~ '/data/search' then --cache(s) inside the query

			if p_query not ilike '%caches%:%{%}%' then
				raise exception 'caches are missing on the query %', p_query;
			end if;

			select array_agg(cac.id)::integer[] , array_agg(a.cache_offsets)
			into v_caches_ids, v_offsets
			from
			(
				select string_to_array( substring(c, '"([^"]*)"') , '/')as cache_uri,
						substring(c, '[[](.*)[]]') as cache_offsets
				from
				(
					select 	( regexp_matches (c, '"[a-fA-F0-9-]{35,52}/[0-9]+/[0-9]+"[ ]*:[ ]*[^"]*', 'g')  )[1] as c
					from
					(
						select substring(p_query, '"?caches"?[ ]*:[ ]*[{][ ]*([^}]*)') as c
					) a
				) a
			) a , l3d."cache" cac
			where
				cac.uuid  = a.cache_uri[1] and
				cac.ext_cache_id0 = a.cache_uri[2] ::integer and
				cac.ext_cache_id1  = a.cache_uri[3] ::integer and
				cac.active ;

			raise notice 'caches %, offsets %', v_caches_ids, v_offsets;
		else
			raise exception 'error on l3d.search_query: wrong uri as argument : %', p_uri;
		end if;


		return query select * from l3d.search_query(p_query,v_caches_ids, v_offsets);
	 	-- -------------------------------------------------------------------------------------------------------------------------------------------------------------
end;
$BODY$;

alter function l3d.search_query(varchar,varchar) owner to i3dhub;



drop function if exists l3d.search_query(varchar,  integer[] ,varchar[]);

create or replace function l3d.search_query(p_query varchar, p_caches_ids integer[], p_offsets varchar[])
RETURNS table(query_result varchar[])
    LANGUAGE 'plpgsql'
AS $BODY$
declare
    -- select * from l3d.search_query('/5b2f17fa79e7cdc3-45d38cf7-4645-b994/0/21422/data/search','{"query":{"select":["edgeHandle"],"conditions":[{"nodeType":"aux"},{"nodeId":998}],"linkDepth":2},"offsets":{"structOffset":0,"auxOffset":0,"shapeInstanceOffset":0}}',false)
    v_fields varchar[];

    v_keys varchar[];
    v_operators varchar[];
    v_values varchar[];

    v_offset integer;

    v_node_type varchar;

    v_offsets varchar[] := p_offsets;
   
begin
    --      raise notice 'search_query: select * from l3d.search_query(''%'',''%'',''%'')',p_query, p_caches_ids, p_offsets;

    select * into v_fields, v_keys, v_operators, v_values from l3d.parse_search_query(p_query, p_caches_ids, p_offsets);

    v_node_type := substring (p_query , '[{][\s]*"?node[tT]ype"?[\s]*:[\s]*"?([a-zA-Z]*)"?[\s]*[}]');

    if v_node_type ilike 'structure' and array_length( p_caches_ids, 1 ) = 1 then

        if v_offsets[1] is null or v_offsets[1] not like '%structOffset=>%' then
            v_offsets[1] := '0,0';
        else
            v_offset := coalesce( coalesce( p_offsets[1]::hstore, 'structOffset=>0'::hstore )->'structOffset', '0' ) ::integer;
            v_offsets[1] := v_offset::varchar || ',' || v_offset::varchar;
        end if;

    end if;
   
	return query 
	with not_or as 
	(
	    select	coalesce( array_agg( k.v ), array[]::varchar[] ) as k, 
				coalesce( array_agg( o.v ), array[]::varchar[] ) as o, 
				coalesce( array_agg( v.v ), array[]::varchar[] ) as v	    
		from
            ( select row_number() over() i, v from unnest (v_keys) v ) k,
            ( select row_number() over() i, v from unnest (v_operators) v ) o,
            ( select row_number() over() i, v from unnest (v_values) v ) v
        where k.i = o.i  and k.i = v.i and o.v not ilike 'or%'	
	), cnd as 
	(
		select not_or.k, not_or.o, not_or.v
		from not_or
        where not exists( select * from  unnest( v_operators ) o where o ilike 'or%' )
		union all        
    	select array_append( not_or.k, k.v ) as k, array_append( not_or.o, substring( o.v, 3 )::varchar ) as o, array_append( not_or.v, v.v ) as v
        from
        	not_or,
            ( select row_number() over() i, v from unnest (v_keys) v ) k,
            ( select row_number() over() i, v from unnest (v_operators) v ) o,
            ( select row_number() over() i, v from unnest (v_values) v ) v
        where k.i = o.i  and k.i = v.i and o.v ilike 'or%'	
	)		
	select l3d.search_query_structure( v_fields, a.k, a.o, a.v, p_caches_ids, v_offsets) st 
	from cnd as a
	where v_node_type ilike 'structure'
	union all
	select l3d.search_query_aux( v_fields, a.k, a.o, a.v, p_offsets, p_caches_ids ) aux
	from cnd as a	
	where v_node_type ilike 'aux' and 'originalFaceID' not ilike all(v_keys) and 'originalFaceID' not ilike all(v_fields)
	union all
	select l3d.search_query_original_face( v_fields, a.k, a.o, a.v, p_offsets, p_caches_ids ) ofc
	from cnd as a
	where v_node_type ilike 'aux' and ( 'originalFaceID' ilike any(v_keys) or 'originalFaceID' ilike any(v_fields) ) ;	
    
end;
$BODY$;

alter function l3d.search_query(varchar, integer[] ,varchar[]) owner to i3dhub;




drop function if exists l3d.parse_search_query( varchar, integer[] ,varchar[]);

create or replace function l3d.parse_search_query(p_query varchar, p_caches_ids integer[] default null, p_offsets varchar[] default null)
returns table(query_fields varchar[], query_condition_keys varchar[], query_condition_operators varchar[], query_condition_values varchar[])
    LANGUAGE 'plpgsql'
AS $BODY$
declare
    v_txt_fields varchar;
    v_txt_conditions varchar;
    v_txt_not_conditions varchar;
    v_txt_or_conditions varchar;

    v_sub_query varchar;

    v_fields varchar[];

    v_keys varchar[];
    v_operators varchar[];
    v_values varchar[];

begin

    select  substring( replace( not_cnd, coalesce( or_cnd, '' ) , '' ) , '[[][\s]*({.*})[\s]*[]]' ) as not_cnd,
            substring( replace( or_cnd, coalesce( not_cnd, '' ) , '' ), '[[][\s]*([{].*[}])[\s]*[]]' ) as or_cnd,
            replace( replace( cnd , coalesce(not_cnd, '') , ''), replace( coalesce( or_cnd, '' ), coalesce(not_cnd , ''), '' ) , '' ) as cnd,
            sel
    into v_txt_not_conditions, v_txt_or_conditions, v_txt_conditions, v_txt_fields
    from
    (
        select  substring(cnd, ',?[\s]*{[\s]*"?not"?[\s]*:[\s]*[[][\s]*{.*}[\s]*[]][\s]*}' ) as not_cnd,
                        substring(cnd, ',?[\s]*{[\s]*"?or"?[\s]*:[\s]*[[][\s]*{.*}[\s]*[]][\s]*}' ) as or_cnd,
                        cnd, sel
        from
        (
            select substring(q, '"?select"?[\s]*:[\s]*[[][\s]*(["][^:]*["])[\s]*[]]') as sel,
                   substring(q, '"?conditions"?[\s]*:[\s]*[[][\s]*([{].*[}])[\s]*[]]') as cnd
            from
            (
                select coalesce( substring(p_query, '"?query"?[\s]*:[\s]*{[\s]*(.*)[\s]*}') , p_query) as q
            ) a
        ) a
    ) a;

    v_sub_query := substring(v_txt_conditions, '{[\s]*"?select"?[\s]*:[\s]*[[][\s]*["].*["][\s]*[]][\s]*,[\s]*"?conditions"?[\s]*:[\s]*[[][\s]*[{].*:.*[}][\s]*[]][\s]*}');

    -- parse fields
    v_fields := regexp_split_to_array( trim( v_txt_fields, '"' ) , '"[\s]*,[\s]*"');


    if v_sub_query is not null then

        declare
            v_sub_query_result varchar;
            v_sub_query_results varchar[] := array[]::varchar[];
            sq record;

        begin

            for sq in ( select * from l3d.search_query(v_sub_query, p_caches_ids, p_offsets) )
            loop

                if array_length( sq.query_result , 1 ) = 1 then

                    v_sub_query_result := sq.query_result[1];

                    if v_sub_query_result !~ '^(([1-9]\d*)|0)((\.|,)\d+)?$' and v_sub_query_result !~ '".*"' then
                            v_sub_query_result := '"' || v_sub_query_result || '"';
                    end if;

                else

                    select '[' || array_to_string( array_agg(a.qr) , ',' ) || ']'
                    into v_sub_query_result
                    from
                    (
                        select case when r ~ '^(([1-9]\d*)|0)((\.|,)\d+)?$' or r ~'^(").*(")$' then r else ( '"' || r || '"') end as qr from unnest(sq.query_result) r
                    ) a;

                end if;

                v_sub_query_results := array_append(v_sub_query_results, v_sub_query_result);

            end loop;

            if array_length( v_sub_query_results , 1 ) = 1 then
                v_sub_query_result := v_sub_query_results[1];

                if v_sub_query ilike '%pmiType%equals%:%FeatureControlFrame%' then
                    v_sub_query_result := '[' || array_to_string( string_to_array( trim( v_sub_query_result, '"'), '-' ), ',' ) || ']';
                end if;

            else
                    v_sub_query_result := '[' || array_to_string( v_sub_query_results , ',' ) || ']';
            end if;

            v_txt_conditions := replace(v_txt_conditions, v_sub_query, v_sub_query_result) ;

        end;
    end if;

    -- parse conditions
    select
        array_agg(k) k,
        array_agg(v) v ,
        array_agg(
            (
                select
                    case when a.is_not_cnd then
                            (array['notEquals', 'lessOrEqualThan', 'greaterOrEqualThan', 'greaterThan', 'lessThan', 'notContains', 'notEqualsAny'])[o.i]
                         when a.is_cs and o.v = 'equals' then
                            'equalss'
                         when a.is_or_cnd then
                            ( 'or' || o.v )
                    else
                         o.v
                    end
                from
                    ( select row_number() over() as i, v from unnest(array['equals', 'greaterThan', 'lessThan', 'lessOrEqualThan', 'greaterOrEqualThan', 'contains', 'equalsAny']) v) o
                where a.o ilike o.v
            )
        ) o
    into v_keys, v_values ,  v_operators
    from
    (
        select k,
            case when v ~ '[[].*[]]' and o = 'equals' then 'equalsAny' else o end as o,
            case
                when v ~ '^([\s]*[[]).*([]][\s]*)$' then
                    (
                        select (  '[' || array_to_string( array_agg( trim(v, '"') ), ',' ) || ']' )
                        from
                        (
                            select regexp_split_to_table( v , '["]?[\s]*,[\s]*["]?') v
                            from
                            (
                               select case when v ~ '^([\s]*["]).*(["][\s]*)$' then substring( v , '[\s]*["](.*)["][\s]*' ) else v end as v
                               from
                               (
                                       select substring( v , '[\s]*[[][\s]*(.*)[\s]*[]][\s]*' ) v
                               ) a
                            ) a
                        ) a
                    )
                when v ~ '^([\s]*["]).*(["][\s]*)$' then
                    substring( v , '[\s]*["](.*)["][\s]*' )
                else
                    substring(v , '[[:graph:]]*' )
            end as v,
            is_cs,
            is_not_cnd,
            is_or_cnd
        from
        (
            select
                case when k ilike 'metadata' then substring(v, '"(.*)"[\s]*,[\s]*"?[[:alpha:]]+"?[\s]*:') else k end as k,
                case when k ilike 'metadata' then substring(v, '"?([[:alpha:]]+)"?[\s]*:') else 'equals' end as o,
                case when k ilike 'metadata' then substring(v, '"?[[:alpha:]]+"?[\s]*:[\s]*(.*)') else v end as v,
                is_cs,
                is_not_cnd,
                is_or_cnd
            from
            (
                select  case when fh is not null then 'faceHandle' else k end,
                        case when fh is not null then fh else replace( v, coalesce( cs, '' ), '' ) end as v,
                        coalesce( substring( cs, ':[\s]*"?([[:alpha:]]*)"?') , 'false' ) ilike 'true' as is_cs,
                        is_not_cnd,
                        is_or_cnd
                from
                (
                    select  substring(c, '{?"?([[:alpha:]]+)"?[\s]*[:]' ) as k ,
                            substring(c, '"?[[:alpha:]]+"?[\s]*[:][\s]*[{]?[\s]*([^}]*)' ) as v     ,
                            substring(c, '[{](.*"face[Ii][Dd]"[\s]*:[^}]*)') as fh,
                            substring(c, '[,]?"?case[sS]ensitive"?[\s]*:[\s]*"?[a-zA-z]*"?[\s]*') as cs,
                            is_not_cnd,
                            is_or_cnd
                    from
                    (
                        select regexp_split_to_table( cnd , '[\s]*[}][\s]*,[\s]*[{][\s]*')  as c, is_not_cnd, is_or_cnd
                        from
                        (
                            select v_txt_conditions as cnd, false as is_not_cnd, false as is_or_cnd
                            union all
                            select v_txt_not_conditions as cnd, true as is_not_cnd, false as is_or_cnd
                            union all
                            select v_txt_or_conditions as cnd, false as is_not_cnd, true as is_or_cnd
                        ) a
                    ) a
                ) a
            ) a
        ) a
    ) a;

    return query select v_fields, v_keys, v_operators, v_values;

end;
$BODY$;

alter function l3d.parse_search_query( varchar, integer[] ,varchar[]) owner to i3dhub;


-- FUNCTION: l3d.search_aux_foriegn_properties(bigint, character varying, character varying[], character varying[], character varying[], integer)

DROP FUNCTION IF EXISTS l3d.search_aux_foriegn_properties(bigint, character varying, character varying[], character varying[], character varying[], integer);

CREATE OR REPLACE FUNCTION l3d.search_aux_foriegn_properties(
	cid bigint,
	auxtype character varying,
	keys character varying[],
	ops character varying[],
	prop_values character varying[],
	off_set integer)
RETURNS TABLE(id bigint, "nodeId" integer, label character varying, "auxType" character varying)
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
    ROWS 1000
AS $BODY$

DECLARE
	rec RECORD;
    rec1 RECORD;
	v_keys varchar[] := null;
BEGIN
	select array_agg(case when key_element ilike any(array['%pmiType', 'nodeId']) then key_element when key_element ilike 'auxProperties.%' then key_element  else ('auxProperties.' || key_element) end)
	into v_keys
    from unnest(keys) as key_element
    where
    key_element not ilike all('{faceHandle,pointsTo,tag%,topoHandle}');


   if not exists(select * from unnest(keys) as key_element where key_element ilike '%pmiType') then

   		v_keys := 		array_append(v_keys, 'pmiType') ;
   		ops :=	 		array_append(ops, 'equals');
   		prop_values :=	array_append(prop_values, auxtype);

   end if;

	return query select * from l3d.search_aux_nodes(v_keys, ops, prop_values, off_set, cid);

END;

$BODY$;

ALTER FUNCTION l3d.search_aux_foriegn_properties(bigint, character varying, character varying[], character varying[], character varying[], integer)
    OWNER TO i3dhub;









-- FUNCTION: l3d.search_l3dnode_metadata(bigint, character varying[], character varying[], character varying[], character varying[], boolean, integer)

DROP FUNCTION IF EXISTS l3d.search_l3dnode_metadata(bigint, character varying[], character varying[], character varying[], character varying[], boolean, integer);

CREATE OR REPLACE FUNCTION l3d.search_l3dnode_metadata(
	cid bigint,
	keys character varying[],
	ops character varying[],
	meta_values character varying[],
	filters character varying[],
	orconditions boolean,
	off_set integer)
RETURNS TABLE(id integer, "nodeId" integer, label character varying, metadata hstore)
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
    ROWS 1000
AS $BODY$

DECLARE
	rec RECORD;
    metaFilters text[]:= ARRAY[]::text[];
    tkn character varying;
BEGIN

	FOR i IN 1..array_length(filters,1) LOOP
    	tkn :=  split_part(filters[i], '.', 2);
    	IF tkn = 'structMetaData' THEN

        	SELECT array_append(metaFilters, replace(split_part(filters[i], '.', 3),'*','%')) INTO metaFilters;


        END IF;
    END LOOP;

	CREATE TEMPORARY TABLE temp_auxIDs (
        	AUXNode_ID bigint,
        	matching int
    ) ON COMMIT PRESERVE ROWS;
	FOR i IN 1..array_length(keys,1) LOOP
    	FOR rec IN (
            		SELECT n.id , md.key, md.value_str as svalue, md.value_num as nvalue
                    FROM l3d.node as n,l3d.metadata as md
                    WHERE md.node_id = n.id
                    AND md.key = keys[i]
					AND n.cache_id = cid
        ) LOOP
        	IF meta_values[i] ~ '^\d+(.\d+)?$' Then
                IF ops[i] = 'lessThan' AND  (rec.nvalue < CAST(meta_values[i] AS DOUBLE PRECISION)) OR rec.svalue < meta_values[i] THEN
                    PERFORM l3d.update_aux_temp_table(rec.id);
                 ELSIF ops[i] = 'equals' AND  rec.nvalue = CAST(meta_values[i] AS DOUBLE PRECISION) OR rec.svalue = meta_values[i] THEN
                    PERFORM l3d.update_aux_temp_table(rec.id);
				ELSIF ops[i] = 'contains' AND  rec.svalue LIKE  replace(meta_values[i],'*','%') OR rec.nvalue IS NULL THEN
                    PERFORM l3d.update_aux_temp_table(rec.id);
                 ELSIF ops[i] = 'greaterThan' AND  rec.nvalue > CAST(meta_values[i] AS DOUBLE PRECISION) OR rec.svalue > meta_values[i] THEN
                    PERFORM l3d.update_aux_temp_table(rec.id);
                 ELSIF ops[i] = 'greaterOrEqualThan' AND  rec.nvalue >= CAST(meta_values[i] AS DOUBLE PRECISION) OR rec.svalue >= meta_values[i] THEN
                    PERFORM l3d.update_aux_temp_table(rec.id);
                 ELSIF ops[i] = 'lessOrEqualThan' AND  rec.nvalue <= CAST(meta_values[i] AS DOUBLE PRECISION) OR rec.svalue <= meta_values[i] THEN
                    PERFORM l3d.update_aux_temp_table(rec.id);
                    END IF;
             ELSE
              IF ops[i] = 'contains' AND  rec.svalue like '%'|| meta_values[i]||'%' AND rec.nvalue IS NULL THEN
            	PERFORM l3d.update_aux_temp_table(rec.id);
            ELSIF ops[i] = 'lessThan' AND  (rec.svalue < meta_values[i]) AND rec.nvalue IS NULL THEN
            	PERFORM l3d.update_aux_temp_table(rec.id);
             ELSIF ops[i] = 'equals' AND  rec.svalue = meta_values[i] AND rec.nvalue IS NULL THEN
            	PERFORM l3d.update_aux_temp_table(rec.id);
             ELSIF ops[i] = 'equals' AND  rec.svalue LIKE  replace(meta_values[i],'*','%') AND rec.nvalue IS NULL THEN
            	PERFORM l3d.update_aux_temp_table(rec.id);
             ELSIF ops[i] = 'equalsAny' AND  rec.svalue LIKE ANY((l3d.array_to_any_operator(meta_values[i]::varchar[]))) AND rec.nvalue IS NULL THEN
            	PERFORM l3d.update_aux_temp_table(rec.id);
             ELSIF ops[i] = 'greaterThan' AND  rec.svalue > meta_values[i] AND rec.nvalue IS NULL THEN
            	PERFORM l3d.update_aux_temp_table(rec.id);
             ELSIF ops[i] = 'greaterOrEqualThan' AND  rec.svalue >= meta_values[i] AND rec.nvalue IS NULL THEN
            	PERFORM l3d.update_aux_temp_table(rec.id);
             ELSIF ops[i] = 'lessOrEqualThan' AND  rec.svalue <= meta_values[i] AND rec.nvalue IS NULL THEN
            	PERFORM l3d.update_aux_temp_table(rec.id);
                END IF;
            END IF;
        END LOOP;
    END LOOP;
    IF  not orConditions THEN
        return QUERY SELECT n.id, n.ext_node_id + off_set as ext_node_id , n.label , hstore(array_agg('metadata.structMetaData.'||md.key), array_agg(COALESCE(md.value_str::text, md.value_num::text)))
        FROM temp_auxIDs, l3d.node as n , l3d.metadata as md
        WHERE n.id =  temp_auxIDs.AUXNode_ID
        AND temp_auxIDs.matching = array_length(keys,1)
        AND md.node_id = n.id
        AND md.key  LIKE ANY(metaFilters::varchar[])
        AND n.cache_id = cid
        GROUP BY n.id, n.ext_node_id,n.label;
    ELSE
    	return QUERY SELECT n.id, n.ext_node_id + off_set as ext_node_id , n.label , hstore(array_agg('metadata.structMetaData.'||md.key), array_agg(COALESCE(md.value_str::text, md.value_num::text)))
        FROM temp_auxIDs, l3d.node as n , l3d.metadata as md
        WHERE n.id =  temp_auxIDs.AUXNode_ID
       -- AND temp_auxIDs.matching = array_length(keys,1)
        AND md.node_id = n.id
        AND md.key  LIKE ANY(metaFilters::varchar[])
        AND n.cache_id = cid
        GROUP BY n.id, n.ext_node_id,n.label;
    END IF;
    DROP TABLE temp_auxIDs;
END;

$BODY$;

ALTER FUNCTION l3d.search_l3dnode_metadata(bigint, character varying[], character varying[], character varying[], character varying[], boolean, integer)
    OWNER TO i3dhub;

-- FUNCTION: l3d.get_topo_by_auxid(bigint, bigint, integer[], refcursor)

DROP FUNCTION IF EXISTS l3d.get_topo_by_auxid(bigint, bigint, integer, integer[], refcursor);

CREATE OR REPLACE FUNCTION l3d.get_topo_by_auxid(
	c_id bigint,
	aux_id bigint,
	shape_instance_offset integer,
    entityType integer[],
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$


BEGIN

OPEN ref FOR
WITH nfi As (
	SELECT an.id, face_instance_id
	FROM l3d.aux_node an , l3d.face_instance_to_aux_node fan
    WHERE an.id = fan.aux_node_id
    AND an.cache_id = c_id
    AND an.ext_aux_node_id = aux_id
),
 nei As (
	SELECT an.id, edge_instance_id
	FROM l3d.aux_node an , l3d.edge_instance_to_aux_node fan
    WHERE an.id = fan.aux_node_id
    AND an.cache_id = c_id
    AND an.ext_aux_node_id = aux_id
),

faceHandles as (SELECT nfi.id, 1 as "type",si.shape_instance_id + shape_instance_offset as "shapeInstanceId" ,f.ext_face_id as "primitiveId"
FROM nfi , l3d.face_instance as fi, l3d.shape_instance as si, l3d.face as f
WHERE fi.shape_instance_id = si.id

AND fi.id = nfi.face_instance_id
AND f.id = fi.face_id),

edgeHandle  as (

SELECT nei.id, nei.edge_instance_id, 2 as "type" , si.shape_instance_id + shape_instance_offset as "shapeInstanceId", e.ext_edge_id as "primitiveId"
 	FROM nei , l3d.edge_instance as ei, l3d.shape_instance as si, l3d.edge as e
	WHERE ei.shape_instance_id = si.id

	AND ei.id = nei.edge_instance_id
	AND e.id = ei.edge_id
 ),
topoHandle as (
SELECT 2 as t, row_to_json((SELECT d1 FROM (SELECT eh.type as "entityType", eh."shapeInstanceId" as "shapeInstanceID", eh."primitiveId" as "entityID") d1))::text as "handle"
FROM edgeHandle as eh

UNION
SELECT 1 as t, row_to_json((SELECT d2 FROM (SELECT fh.type as "entityType", fh."shapeInstanceId" as "shapeInstanceID", fh."primitiveId" as "entityID") d2))::text as "handle"

FROM faceHandles as fh)

SELECT "handle" as "topoHandle"
FROM topoHandle
WHERE topoHandle.t = ANY(entityType::integer[])
;



RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_topo_by_auxid(bigint, bigint, integer,integer[], refcursor)
    OWNER TO i3dhub;






DROP FUNCTION IF EXISTS l3d.search_text(bigint, character varying);

CREATE OR REPLACE FUNCTION l3d.search_text(
	cid bigint,
	keyword character varying)
RETURNS TABLE(suggestion character varying, type character varying)
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
    ROWS 1000
AS $BODY$

DECLARE
	rec RECORD;
BEGIN
	CREATE TEMPORARY TABLE suggestions(
        	"word" VARCHAR,
        	"type" VARCHAR
    ) ON COMMIT PRESERVE ROWS;

    INSERT INTO suggestions SELECT "Label", 'auxProperties.Label'
    FROM l3d.aux_node as an
    WHERE lower(an."Label") LIKE '%'|| lower(keyword) || '%'
    AND an.cache_id = cid LIMIT 5;

    INSERT INTO suggestions SELECT name, 'auxProperties.pmiType'
    FROM l3d.aux_node_type as ant
    WHERE lower(ant.name) LIKE '%'|| lower(keyword) || '%'
    LIMIT 5;

    INSERT INTO suggestions SELECT av.value_str, 'auxProperties.'|| avt.name
    FROM l3d.aux_value_name as avt, l3d.aux_value as av
    WHERE avt.id = av.aux_value_name_id
    AND lower(av.value_str) LIKE '%'|| lower(keyword) || '%'
    LIMIT 5;

    INSERT INTO suggestions SELECT aat.value, 'auxAttributes.'|| aat.key
    FROM l3d.aux_attributes as aat
    WHERE lower(aat.value) LIKE '%'|| lower(keyword) || '%'
    LIMIT 5;

    INSERT INTO suggestions SELECT md.value_str, 'structMetaData.'|| md.key
    FROM l3d.metadata as md
    WHERE lower(md.value_str) LIKE '%'|| lower(keyword) || '%'
    LIMIT 5;

	return QUERY SELECT *  FROM suggestions;
    DROP TABLE suggestions;
END;

$BODY$;

ALTER FUNCTION l3d.search_text(bigint, character varying)
    OWNER TO i3dhub;



-- FUNCTION: l3d.search_aux_attributes(bigint, character varying, character varying[], character varying[], character varying[], integer)

DROP FUNCTION IF EXISTS l3d.search_aux_attributes(bigint, character varying, character varying[], character varying[], character varying[], integer);

CREATE OR REPLACE FUNCTION l3d.search_aux_attributes(
	cid bigint,
	auxtype character varying,
	keys character varying[],
	ops character varying[],
	prop_values character varying[],
	off_set integer)
RETURNS TABLE(id bigint, "nodeId" integer, label character varying , "auxType" character varying)
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
    ROWS 1000
AS $BODY$
declare
  	-- SELECT * FROM l3d.search_aux_attributes(4846, '*', array['pmiType','Description'],  array['contains','equalsAny'], array['Lin', '{%Cen%,%Lin%;%Not%}'],0);
	v_keys varchar[] := null;
BEGIN

	select array_agg(case when key_element ilike 'auxAttributes.%' then key_element when key_element ilike '%pmiType' then 'auxProperties.pmiType' else 'auxAttributes.' || key_element end) into v_keys
	from unnest(keys) as key_element
	where key_element not ilike all('{nodeId,faceHandle,pointsTo,tag%,topoHandle}');

	return query select * from l3d.search_aux_nodes(v_keys, ops, prop_values, off_set, cid);
END;

$BODY$;

ALTER FUNCTION l3d.search_aux_attributes(bigint, character varying, character varying[], character varying[], character varying[], integer)
    OWNER TO i3dhub;







-- FUNCTION: l3d.get_aux_by_faceinstance(bigint, bigint, bigint, character varying[], character varying[], character varying[], integer)

DROP FUNCTION IF EXISTS l3d.get_aux_by_faceinstance(bigint, bigint, bigint, character varying[], character varying[], character varying[], integer);

CREATE OR REPLACE FUNCTION l3d.get_aux_by_faceinstance(p_cache_id bigint, p_face_id bigint, p_shape_instance_id bigint, p_keys character varying[], p_relations character varying[], p_values character varying[], p_offset integer DEFAULT 0)
RETURNS TABLE("nodeId" integer)
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$

declare
 -- 	SELECT * from l3d.get_aux_by_faceInstance(15240,49960,1, array['faceHandle','auxProperties.pmiType','auxProperties.type'], Array['equals','contains', 'contains'], ARRAY['}','Dime', 'lin']);
	r_aux record;
	r_nodes integer[] = array[]::integer[] ;
begin

	select array_agg(fia.aux_node_id) into r_nodes
	from
		face fac,
		shape_instance shi,
		face_instance fin,
		face_instance_to_aux_node fia
	where
		fac.cache_id = p_cache_id					and
		fac.ext_face_id = p_face_id 				and

		shi.cache_id = p_cache_id 					and
		shi.shape_instance_id = p_shape_instance_id	and

		fin.cache_id = p_cache_id					and
		fin.shape_instance_id = shi.id				and
		fin.face_id = fac.id						and

		fia.face_instance_id = fin.id;

	return query
		select a.nodeid from l3d.search_aux_nodes(p_keys, p_relations, p_values, p_offset, 0, r_nodes) a;
END;
$BODY$;

ALTER FUNCTION l3d.get_aux_by_faceinstance(bigint, bigint, bigint, character varying[], character varying[], character varying[], integer)
    OWNER TO i3dhub;




-- FUNCTION: l3d.collect_and_filter_search_results(bigint, character varying, character varying[], character varying[], character varying[], character varying[], boolean, integer)
-- DROP FUNCTION l3d.collect_and_filter_search_results(bigint, character varying, character varying[], character varying[], character varying[], character varying[], boolean, integer);

DROP FUNCTION IF EXISTS l3d.collect_and_filter_search_results(bigint, character varying, character varying[], character varying[], character varying[], character varying[], integer, integer);

CREATE OR REPLACE FUNCTION l3d.collect_and_filter_search_results(
	cid bigint,
	auxtype character varying,
	keys character varying[],
	ops character varying[],
	prop_values character varying[],
	filters character varying[],
	searchprops integer,
	off_set integer)
RETURNS TABLE(id bigint, "nodeId" integer, label character varying, "metadata.auxProperties.pmiType" character varying, metadata hstore, "pointsTo" varchar)
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
    ROWS 100000
AS $BODY$
DECLARE
	propertiesFilters text[]:= ARRAY[]::text[];
    attributesFilters text[] := ARRAY[]::text[];
    tkn character varying;
    rec RECORD;
    tmp_hstore hstore;
	filter_key character varying;

	json_invalid_characters hstore;
begin
	--check if pointsTo need to be included in the results
	IF (SELECT 'pointsTo' = ANY(filters)) THEN
		select array_agg(elements) from
   				(select unnest(filters)
				except
   				select unnest('{pointsTo}'::varchar[])) t (elements)
		INTO filters;
		RETURN QUERY SELECT d.id, d."nodeId",d.label, d."metadata.auxProperties.pmiType", d.metadata, ('[' || string_agg( an.ext_aux_node_id::text ,',') || ']')::varchar
		FROM l3d.collect_and_filter_search_results(cid,auxtype,keys,ops,prop_values,filters,searchprops,off_set) as d
		LEFT JOIN l3d.aux_node_to_aux_node as ansp  on ansp.pred_id = d.id

		LEFT JOIN l3d.aux_node as an on ansp.succ_id = an.id
	         GROUP BY d.id, d."nodeId",d.label, d."metadata.auxProperties.pmiType", d.metadata;
	ELSE --start searching
	-- split the filter into properties and attribute filters
    FOR i IN 1..array_length(filters,1) LOOP
    	tkn :=  split_part(filters[i], '.', 2);
		IF filters[i] != 'nodeId' and filters[i] != 'label' and filters[i] != 'Label' THEN
			filter_key :=  substring( filters[i], length('metadata.'||tkn || '.')+1, (length(filters[i]) - length('metadata'||tkn || '.') ) );

		END IF;
    	IF tkn = 'auxProperties' THEN
        	IF filter_key = 'pmiType' Then
				SELECT array_append(propertiesFilters, replace(filter_key,'pmiType','%')) INTO propertiesFilters;
			ELSE
        		SELECT array_append(propertiesFilters, replace(filter_key,'*','%')) INTO propertiesFilters;
			END IF;

       	ELSIF tkn = 'auxAttributes' THEN
        	SELECT array_append(attributesFilters, replace(filter_key,'*','%')) INTO attributesFilters;

        END IF;
    END LOOP;

--   raise notice 'l3d.search_aux_nodes(''%'', ''%'', ''%'', %, %)', keys, ops, prop_values, off_set, cid;

	return query
	select	r.id,
			r."nodeId",
	   		r.label,
			r."auxType",
			hstore(
				'metadata.array',
				(
					select array_to_string(array_agg(v), ',')
					from
					(
						-- offending characters that could prevent parsing in json
						select regexp_replace(regexp_replace(regexp_replace(a.v, E'\\u0009+', '\\b', 'g'), E'\\u000C+', '\\f', 'g'), E'\\u0009+', '\\t', 'g') as v
						from
						(
							select regexp_replace(regexp_replace(regexp_replace(a.v, E'\\u000A+', '\\n', 'g'), E'\\u000D+', '\\r', 'g'), E'\\u0000+', '\\0', 'g') as v
							from
							(
								select replace(coalesce(v, '"undefined"'), '@@^^@@', case when v like '{%}' then '' else '"' end) as v
								from unnest(r.metadata->(filters)) v
							) a
						) a

					) a
				)) as metadata,
			'{}'::varchar as "pointsTo"
	from
	(
		select  r.id, r."nodeId", r.label, r."auxType",

				case when coalesce(array_upper(propertiesFilters, 1), 0) > 0 then
					case when exists(select * from unnest(filters) f where f ilike 'metadata.auxProperties.*') then
							hstore('metadata.auxProperties.*', '{'|| auxProperties || '}')
					else
							hstore('')
					end || replace(auxProperties, '":"', '"=>"')::hstore
				else
					hstore('')
				end ||
				case when coalesce(array_upper(attributesFilters, 1), 0) > 0 then
					case when exists(select * from unnest(filters) f where f ilike 'metadata.auxAttributes.*') then
							hstore('metadata.auxAttributes.*', '{'|| auxAttributes || '}')
					else
							hstore('')
					end || replace(auxAttributes, '":"', '"=>"')::hstore
				else
					hstore('')
				end ||
				hstore('nodeId',  r."nodeId"::text ) || labels as metadata
		from
		(
			SELECT r.*,
					(
						select hstore(array_agg(k), array_agg('@@^^@@'|| v || '@@^^@@')) as labels
						from
						(
							select k , replace(trim(v, '"'), '"', '\"') as v
							from
							(
								select 'label' as k , r.label as v union all select 'Label' as k , r.label as v
							) a
						) a
					),
				  (
--				  	select coalesce(array_to_string(array_agg('"metadata.auxProperties.' || k || '":"' || regexp_replace(v, E'[\\f\\u000B\\u0085\\u2028\\u2029]+', ' ', 'g') || '"' ) , ','), '') auxProperties
				  	select coalesce(array_to_string(array_agg('"metadata.auxProperties.' || k || '":"@@^^@@' || v || '@@^^@@"' ) , ','), '') auxProperties
				  	from
				  	(
						select avn.name as k, coalesce(replace(replace(trim(av.value_str, '"'), '\', '\\'), '"', '\"'),  av.value_num::varchar) as v
						from aux_value av , aux_value_name avn
					  	where
						 avn.name like any(propertiesFilters) and coalesce(array_upper(propertiesFilters, 1), 0) > 0 and av.aux_node_id = r.id and av.aux_value_name_id  = avn.id
					  	union all
						select k, v
						from
						(
							select 'pmiType'::text as k,  r."auxType" as  v
							where
							coalesce(array_upper(propertiesFilters, 1), 0) > 0 and exists(select * from unnest(propertiesFilters) f where f in ('%', 'pmiType'))
						) a
				  	) a
				  ),
				  (
				  	select coalesce(array_to_string(array_agg( '"metadata.auxAttributes.' || k || '":"@@^^@@' || v || '@@^^@@"'), ',' ), '') auxAttributes
				  	from
				  	(
				  		select att.key as k, replace(replace(trim(att.value, '"'), '\', '\\'), '"', '\"') as v
					  	from aux_attributes att
					  	where
					  		att.aux_node_id = r.id and att.key like any(attributesFilters) and coalesce(array_upper(attributesFilters, 1), 0) > 0
				  	) a
				  )
			from
			(
				select a.id, a.nodeid as "nodeId", a.label, a.auxtype as "auxType"
				from l3d.search_aux_nodes(keys, ops, prop_values, off_set, cid)  a
				where
					exists(select * from unnest(keys) k where k like any(array['auxProperties.%', 'auxAttributes.%']))
			) r
		) r
	) r;

	END IF;--end search
END;
$BODY$;

ALTER FUNCTION l3d.collect_and_filter_search_results(bigint, character varying, character varying[], character varying[], character varying[], character varying[], integer, integer)
    OWNER TO i3dhub;





-- FUNCTION: l3d.collect_and_filter_search_results(bigint, character varying, character varying[], character varying[], character varying[], character varying[], boolean, integer)
-- This function is being tested as replacement to l3d.collect_and_filter_search_results
-- DROP FUNCTION l3d.collect_and_filter_search_results(bigint, character varying, character varying[], character varying[], character varying[], character varying[], boolean, integer);

DROP FUNCTION IF EXISTS l3d.collect_and_filter_nodeid_search_results(bigint, character varying, character varying[], character varying[], character varying[], character varying[], boolean, integer);

CREATE OR REPLACE FUNCTION l3d.collect_and_filter_nodeid_search_results(
	cid bigint,
	auxtype character varying,
	keys character varying[],
	ops character varying[],
	prop_values character varying[],
	filters character varying[],
	searchprops boolean,
	off_set integer)
RETURNS TABLE("nodeId" integer)
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
    ROWS 100000
AS $BODY$

BEGIN

  	-- create temporary tables
    IF searchProps THEN
    	CREATE TEMPORARY TABLE  results AS
    		SELECT * FROM l3d.search_aux_foriegn_properties(cid, auxType, keys, ops, prop_values,0)
            UNION
            SELECT * FROM l3d.search_aux_attributes(cid, auxType, keys, ops, prop_values,0);
            ANALYZE results;
  	ELSE
    	CREATE TEMPORARY TABLE  results AS
    		SELECT * FROM l3d.search_aux_attributes(cid, auxType, keys, ops, prop_values,0);
    END IF;

    ANALYZE results;
    RETURN QUERY SELECT results."nodeId" FROM results;

    DROP TABLE results;
END;

$BODY$;

ALTER FUNCTION l3d.collect_and_filter_nodeid_search_results(bigint, character varying, character varying[], character varying[], character varying[], character varying[], boolean, integer)
    OWNER TO i3dhub;










-- FUNCTION: l3d.compare_caches(bigint, bigint, refcursor)
DROP FUNCTION IF EXISTS l3d.compare_caches(bigint, bigint, refcursor);

CREATE OR REPLACE FUNCTION l3d.compare_caches(
	cid1 bigint,
	cid2 bigint,
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$

BEGIN

    OPEN ref FOR SELECT avn.name as name ,an1.cache_id as "cid1" ,an1.aux_node_id as "nodeId1",av1.value_str as value1, av1.value_num as value1, an2.cache_id as "cid2", an2.aux_node_id as "nodeId2",  av2.value_str as value2, av2.value_num as value2
    FROM l3d.aux_value as av1 ,l3d.aux_value as av2 , l3d.aux_value_name as avn , l3d.aux_node as an1 , l3d.aux_node as an2
    WHERE av1.aux_value_name_id = avn.id
    AND av2.aux_value_name_id = avn.id
    AND an1.id = av1.aux_node_id
    AND an2.id = av2.aux_node_id
    AND an1.cache_id = cid1
    AND an2.cache_id = cid2
    AND an1."Label" = an2."Label"
    AND (av1.value_str <> av2.value_str or av1.value_num <> av2.value_num ) ;

    RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.compare_caches(bigint, bigint, refcursor)
    OWNER TO i3dhub;





-- hlepr function

DROP FUNCTION IF EXISTS l3d.array_to_any_operator(character varying[]);

CREATE OR REPLACE FUNCTION l3d.array_to_any_operator(
	invalues character varying[])
RETURNS character varying[]
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE

AS $BODY$
DECLARE
	out_array text[]:= ARRAY[]::text[];
BEGIN

	For i in 1..array_length(invalues,1) LOOP
    	SELECT array_append(out_array, replace(invalues[i],'*','%')) INTO out_array;
    END LOOP;

    RETURN out_array;
END;

$BODY$;

ALTER FUNCTION l3d.array_to_any_operator(character varying[])
    OWNER TO i3dhub;

--
-- helper procedure to get the type (string or numeric) of aux property
--

DROP FUNCTION IF EXISTS l3d.get_auxProp_type(character varying);

CREATE OR REPLACE FUNCTION l3d.get_auxProp_type(
	prop_name character varying)
RETURNS character varying
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$

	DECLARE t character varying;
BEGIN


	SELECT
		CASE
			WHEN value_str IS NULL AND value_num IS NOT NULL THEN
				'double precision'
			ELSE
				'text'
		END
	FROM l3d.aux_value as av, l3d.aux_value_name as avn
	WHERE av.aux_value_name_id = avn.id
	AND avn.name = prop_name

	INTO t;
	IF t IS NULL THEN
		RETURN 'text';
	ELSE
		RETURN t;
	END IF;
    END;

$BODY$;

ALTER FUNCTION l3d.get_auxProp_type(character varying)
    OWNER TO i3dhub;



--
-- insert tags procedure
--

DROP FUNCTION IF EXISTS l3d.ins_tagsnames(character varying[]);

CREATE OR REPLACE FUNCTION l3d.ins_tagsnames(
	p_names character varying[]
	)
RETURNS integer[]
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$
  DECLARE ids integer[];
  DECLARE id integer;
 BEGIN
 	FOR i IN 1..array_length(p_names,1) LOOP
		-- check if the name exits
		SELECT tn.id FROM l3d.tag_name as tn WHERE name = p_names[i] INTO id;
		IF FOUND THEN
			SELECT array_append(ids, id) INTO ids;
		ELSE
			INSERT INTO l3d.tag_name(name) VALUES (p_names[i]) RETURNING l3d.tag_name.id into id;
			SELECT array_append(ids, id) INTO ids;
		END IF;
	END LOOP;
	RETURN ids;
 END;
$BODY$;

ALTER FUNCTION l3d.ins_tagsnames(character varying[])
    OWNER TO i3dhub;

-- FUNCTION: l3d.calculate_datastats(bigint)

DROP FUNCTION IF EXISTS l3d.calculate_datastats(bigint);

CREATE OR REPLACE FUNCTION l3d.calculate_datastats(

	cid bigint
	)
RETURNS void
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$


BEGIN
--create temp table to collect stats
CREATE TEMPORARY TABLE totalStats (
        	"type" varchar,
        	count int,
			"stats" varchar
    ) ON COMMIT PRESERVE ROWS;

--count auxtypes
INSERT INTO totalStats
	SELECT d.name, sum (d."count") as "count", '"statistic":[]' FROM (

	SELECT ant.name as name , 0 as "count"
	FROM l3d.aux_node_type as ant

	UNION ALL

	SELECT ant.name as name, COUNT (an.id) as count
	FROM l3d.aux_node as an, l3d.aux_node_type as ant
	WHERE an.aux_node_type_id = ant.id
	AND an.cache_id = cid
	GROUP BY ant.name) d group by d.name;


--collect modelview stats
UPDATE totalStats SET "stats" = '"statistic":'||COALESCE((
SELECT array_to_json(array_agg(row_to_json(d))) FROM (
SELECT t.ext_aux_node_id as "modelViewId", array_agg(t.stats) as statistic FROM(
SELECT an1.ext_aux_node_id,('{"name":"'||ant.name ||'" ,"count": '||(COUNT ( asp.succ_id))::text || '}')::json as stats
	FROM l3d.aux_node_to_aux_node as asp , l3d.aux_node as an, l3d.aux_node_type as ant,l3d.aux_node_type as ant1,
	l3d.aux_node as an1

    WHERE asp.succ_id = an.id
    AND an.aux_node_type_id = ant.id
	 AND an1.aux_node_type_id = ant1.id
  	AND asp.pred_id = an1.id
	AND ant1.name = 'ModelView'
	AND an.cache_id = cid

    GROUP BY an1.ext_aux_node_id, ant.name
	)t GROUP BY ext_aux_node_id

	)d

),'[]') WHERE type ='ModelView';


--collect dimension subtypes stats
UPDATE totalStats SET "stats" = '"subTypes":'||COALESCE((
	SELECT  array_to_json(array_agg(row_to_json(t)))  FROM
	(
		SELECT  value_str as "name", COUNT (value_str) as count
		FROM l3d.aux_value as av, l3d.aux_node as an
		WHERE  av.aux_node_id= an.id
		AND an.aux_node_type_id = (SELECT id FROM l3d.aux_node_type WHERE name = 'Dimension')
		AND av.aux_value_name_id = (SELECT id FROM l3d.aux_value_name WHERE name = 'type')
		AND an.cache_id = cid
		AND value_num is NULL
		GROUP BY value_str
	)t
),'[]') WHERE type='Dimension';

--collect dimension fits stats
UPDATE totalStats SET "stats" = "stats"||',"fits":'||COALESCE((
	SELECT  array_to_json(array_agg(row_to_json(t)))  FROM
	(
		SELECT  value_str as "name", COUNT (value_str) as count
		FROM l3d.aux_value as av, l3d.aux_node as an
		WHERE  av.aux_node_id= an.id
		AND an.aux_node_type_id = (SELECT id FROM l3d.aux_node_type WHERE name = 'Dimension')
		AND av.aux_value_name_id = (SELECT id FROM l3d.aux_value_name WHERE name = 'fit')
		AND an.cache_id = cid
		AND value_num is NULL
		GROUP BY value_str
	)t
),'[]')  WHERE type='Dimension';


--collect SurfaceFinish roughness
UPDATE totalStats SET "stats" = '"roughness":'||COALESCE((
	SELECT  array_to_json(array_agg(row_to_json(t)))  FROM
	(
		SELECT  value_str as "name", COUNT (value_str) as count
		FROM l3d.aux_value as av, l3d.aux_node as an
		WHERE  av.aux_node_id= an.id
		AND an.aux_node_type_id = (SELECT id FROM l3d.aux_node_type WHERE name = 'SurfaceFinish')
		AND av.aux_value_name_id = (SELECT id FROM l3d.aux_value_name WHERE name = 'maxRoughness')
		AND an.cache_id = cid
		AND value_num is NULL
		GROUP BY value_str
	)t
),'[]') WHERE type='SurfaceFinish';


--collect FeatureControlFrame subtypes
UPDATE totalStats SET "stats" = '"subTypes":'||COALESCE((
	SELECT  array_to_json(array_agg(row_to_json(t)))  FROM
	(
		SELECT  value_str as "name", COUNT (value_str) as count
		FROM l3d.aux_value as av, l3d.aux_node as an
		WHERE  av.aux_node_id= an.id
		AND an.aux_node_type_id = (SELECT id FROM l3d.aux_node_type WHERE name = 'FeatureControlFrame')
		AND av.aux_value_name_id = (SELECT id FROM l3d.aux_value_name WHERE name = 'characteristic')
		AND an.cache_id = cid
		AND value_num is NULL
		GROUP BY value_str
	)t
),'[]') WHERE type='FeatureControlFrame';


INSERT INTO l3d.data_stats
	SELECT cid, array_to_json(array_agg(('{"name":"'||type||'","count":'|| count::varchar||','||stats||'}')::json)) FROM totalStats ;

DROP TABLE totalStats;



END;

$BODY$;

ALTER FUNCTION l3d.calculate_datastats(bigint)
    OWNER TO i3dhub;


-- FUNCTION: l3d.get_datastats(bigint, refcursor)

DROP FUNCTION IF EXISTS l3d.get_datastats(bigint, refcursor);

CREATE OR REPLACE FUNCTION l3d.get_datastats(
	cid bigint,
	ref refcursor)
RETURNS refcursor
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE
AS $BODY$

BEGIN
OPEN ref FOR SELECT stats::varchar FROM l3d.data_stats WHERE cache_id = cid;
RETURN ref;
END;

$BODY$;

ALTER FUNCTION l3d.get_datastats(bigint, refcursor)
    OWNER TO i3dhub;

