--
-- Script to create tables to store indexL3d data
-- Note: Make sure you are connected to instant3Dhub when executing this script
--

--
-- Table l3d.index_source
--
CREATE TABLE l3d.index_source
(
    uri character varying,
    url character varying,
    auth_url character varying,
    mime_type character varying[],
    property character varying,
    e_tag character varying,
    cache_id bigint,
    CONSTRAINT pk_index_source_id PRIMARY KEY (cache_id),
    CONSTRAINT fk_source_cache_id FOREIGN KEY (cache_id)
        REFERENCES l3d.cache(id) MATCH SIMPLE
        ON UPDATE NO ACTION
        ON DELETE NO ACTION
)
WITH (
    OIDS = FALSE
);

ALTER TABLE l3d.index_source
    OWNER to i3dhub;


--
-- Table l3d.index_cache
--
CREATE TABLE l3d.index_cache
(
    
    url character varying,
    state character varying,
    transcoder_group character varying,
    bytes integer,
    cache_id bigint,
	l3dgen_version character varying,
	data_layout_version integer,
	original_request json,
    CONSTRAINT pk_index_cache_id PRIMARY KEY (cache_id),
    CONSTRAINT fk_cache_id FOREIGN KEY (cache_id)
        REFERENCES l3d.cache(id) MATCH SIMPLE
        ON UPDATE NO ACTION
        ON DELETE NO ACTION
)
WITH (
    OIDS = FALSE
);

ALTER TABLE l3d.index_cache
    OWNER to i3dhub;


--
-- Table l3d.index_stats
--
CREATE TABLE l3d.index_stats
(
    
    queued_date timestamp with time zone,
    num_reads integer,
    last_modified_date timestamp with time zone,
    last_read_date timestamp with time zone,
    processing_start_date timestamp with time zone,
    processing_end_date timestamp with time zone,
    cache_id bigint,
    CONSTRAINT pk_index_stats_id PRIMARY KEY (cache_id),
    CONSTRAINT fk_cache_id FOREIGN KEY (cache_id)
        REFERENCES l3d.cache(id) MATCH SIMPLE
        ON UPDATE NO ACTION
        ON DELETE NO ACTION
)
WITH (
    OIDS = FALSE
);

ALTER TABLE l3d.index_stats
    OWNER to i3dhub;



--
-- Table l3d.index_stats
--
CREATE TABLE l3d.index_properties
(
    
    properties json,
    cache_id bigint,
    CONSTRAINT pk_index_properties_id PRIMARY KEY (cache_id),
    CONSTRAINT fk_cache_id FOREIGN KEY (cache_id)
        REFERENCES l3d.cache(id) MATCH SIMPLE
        ON UPDATE NO ACTION
        ON DELETE NO ACTION
)
WITH (
    OIDS = FALSE
);

ALTER TABLE l3d.index_properties
    OWNER to i3dhub;



--
-- Add indices
--

-- Index: indeX_index_cache_cache_id

-- DROP INDEX l3d."indeX_index_cache_cache_id";

CREATE INDEX "indeX_index_cache_cache_id"
    ON l3d.index_cache USING btree
    (cache_id)
    TABLESPACE pg_default;


-- Index: index_index_properties_cache_id

-- DROP INDEX l3d.index_index_properties_cache_id;

CREATE INDEX index_index_properties_cache_id
    ON l3d.index_properties USING btree
    (cache_id)
    TABLESPACE pg_default;


-- Index: index_index_source_id

-- DROP INDEX l3d.index_index_source_id;

CREATE INDEX index_index_source_cache_id
    ON l3d.index_source USING btree
    (cache_id)
    TABLESPACE pg_default;


-- Index: index_index_stats_cache_id

-- DROP INDEX l3d.index_index_stats_cache_id;

CREATE INDEX index_index_stats_cache_id
    ON l3d.index_stats USING btree
    (cache_id)
    TABLESPACE pg_default;

