------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------
--
-- This file contains all stored procedures related to searching in struct nodes and its assosiated data (metadata)
--
--
------------------------------------------------------------------------------------------------------------------
------------------------------------------------------------------------------------------------------------------

-- search for extensions in l3d schema
SET search_path TO l3d, extensions;

--
-- Function l3d.search_struct_node_id
-- search structure nodes by node_id 
--

DROP FUNCTION IF EXISTS l3d.search_struct_node_id(hstore, character varying, integer);

CREATE OR REPLACE FUNCTION l3d.search_struct_node_id(
	cids hstore,
	op character varying,
	node_id integer)
    RETURNS TABLE(id integer, "nodeId" integer, cache_id integer, label character varying) 
    LANGUAGE 'plpgsql'

    COST 100
    VOLATILE 
    ROWS 1000
AS $BODY$

BEGIN

    RETURN QUERY 
	SELECT n.id, 
	n.ext_node_id
	as "nodeId",
	n.cache_id
	, n.label as "Label"
	FROM l3d.node as n  
	WHERE n.cache_id = ANY(akeys(cids)::bigint[])
	AND n.ext_node_id = node_id;
    
 

END;

$BODY$;

ALTER FUNCTION l3d.search_struct_node_id(hstore, character varying, integer)
    OWNER TO i3dhub;


--
-- Function l3d.search_struct_label
-- search labels of structure nodes 
--

DROP FUNCTION IF EXISTS l3d.search_struct_label(hstore, character varying, character varying);

CREATE OR REPLACE FUNCTION l3d.search_struct_label(
	cids hstore,
	op character varying,
	labelstr character varying)
RETURNS TABLE(id integer, "nodeId" integer,"cache_id" integer, "label" character varying) 
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE 
    ROWS 1000
AS $BODY$

declare
	-- select * from l3d.search_struct_label('29458 => 29458', 'contains', 'assembly')
	v_label_str varchar := replace(trim(labelStr,'"'), '*','%');
	v_labels varchar[] ;
	v_cache_ids bigint[] := akeys(cids)::bigint[];
begin
	
	if op = 'equalsAny' then
		v_labels := v_label_str ::varchar[];
	elsif op = 'contains' then
		v_labels[1] := '%' || v_label_str|| '%';
	else
		v_labels[1] := v_label_str;
	end if;
	
	RETURN QUERY 		
	SELECT n.id, n.ext_node_id as "nodeId", n.cache_id, n.label as "Label"		
	FROM 		
		node as n 
	WHERE 
	n.cache_id in (select c_i from unnest(v_cache_ids) as c_i) and 
	exists
	(
		select * from unnest(v_labels) as l where n.label ilike l
	);

	
END;

$BODY$;

ALTER FUNCTION l3d.search_struct_label(hstore, character varying, character varying)
    OWNER TO i3dhub;



-- FUNCTION: l3d.search_struct_tags(hstore, character varying, character varying, character varying)

-- DROP FUNCTION l3d.search_struct_tags(hstore, character varying, character varying, character varying);

DROP FUNCTION IF EXISTS l3d.search_struct_tags(hstore, character varying, character varying, character varying);

CREATE OR REPLACE FUNCTION l3d.search_struct_tags(
	cids hstore,
	t_key character varying,
	op character varying,
	t_value character varying)
RETURNS TABLE(id integer, "nodeId" integer,cache_id integer, "label" character varying) 
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE 
    ROWS 1000
AS $BODY$

DECLARE
 query character varying;
BEGIN

	
	-- build thee query based on the operator
	
	--offset handling 
	
	query:= $$ SELECT   n. id, ext_node_id as "nodeId",n.cache_id as "cache_id", n.label as "label"
	FROM l3d.node as n , l3d.tags_values as tv , l3d.tag_name as tn
	WHERE n.id = tv.node_id
	AND tv.tag_name_id = tn.id
	AND n.cache_id = ANY('{$$ || array_to_string(akeys(cids),',','0') || $$}'::bigint[])
	AND tn.name = '$$ || t_key || $$'$$ ;
	

		IF op = 'contains'  THEN
			query:= query || $$ AND tv.value like '%$$|| t_value ||$$%' $$; 
		ELSIF op = 'lessThan'  THEN
			query:= query || $$ AND tv.value < '$$ || t_value || $$' $$; 
		ELSIF op = 'equals'  THEN
			query:= query || $$ AND tv.value LIKE  replace('$$|| t_value || $$','*','%')  $$;
		ELSIF op = 'equalsAny'  THEN
			query:= query || $$ AND tv.value LIKE ANY((l3d.array_to_any_operator('$$ || t_value || $$'::varchar[])))  $$;   
		ELSIF op = 'greaterThan'  THEN
			query:= query || $$ AND tv.value > '$$ || t_value || $$' $$;    
		ELSIF op = 'greaterOrEqualThan' THEN
			query:= query || $$ AND tv.value >= '$$ || t_value || $$' $$; 
		ELSIF op = 'lessOrEqualThan' THEN
			query:= query || $$ AND tv.value <= '$$ || t_value || $$' $$; 
		END IF;
       
	RAISE NOTICE 'query % ',query;
	RETURN QUERY EXECUTE query;	

END;

$BODY$;

ALTER FUNCTION l3d.search_struct_tags(hstore, character varying, character varying, character varying)
    OWNER TO i3dhub;




-- DROP FUNCTION IF EXISTS search_ir3d_struct_metadata(hstore,character varying,character varying,character varying);

--
-- Function l3d.search_struct_metadata
-- search for metadata based on {key} op {value}
--

DROP FUNCTION IF EXISTS l3d.search_struct_metadata(hstore, character varying, character varying, character varying);

CREATE OR REPLACE FUNCTION l3d.search_struct_metadata(
	cids hstore,
	md_key character varying,
	op character varying,
	md_value character varying
	)
RETURNS TABLE(id integer, "nodeId" integer,cache_id integer, "label" character varying) 
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE 
    ROWS 1000
AS $BODY$

DECLARE	
	v_value varchar :=  replace(md_value, '*', '%');
BEGIN
	
	return QUERY	
	SELECT   nod.id, nod.ext_node_id as "nodeId",nod.cache_id, nod.label as "label"
	FROM l3d.node as nod , l3d.metadata as met
	where
	nod.cache_id = ANY(akeys(cids)::integer[]) 		and
	nod.id = met.node_id					and
	met.key ilike md_key					and
	(
		(				  	
			met.value_str is not null and
			(
				(op = 'contains' and met.value_str like v_value) 								or
				(op in ('equals', 'greaterOrEqualThan', 'lessOrEqualThan') and met.value_str = md_value) 			or				
				(op in ('lessThan', 'lessOrEqualThan') and met.value_str < md_value)						or
				(op in ('greaterThan', 'greaterOrEqualThan') and met.value_str > md_value)					or
				(op = 'equalsAny' and met.value_str like ANY(v_value::varchar[]))
			)							
		) or	
		(
			md_value ~ '^\d+(.\d+)?$'	and
			met.value_num is not null	and
			(				  	
				(op in ('equals', 'greaterOrEqualThan', 'lessOrEqualThan') and met.value_num = CAST(md_value AS numeric)) 	or
				(op in ('lessThan', 'lessOrEqualThan') and met.value_num < CAST(md_value AS numeric))				or
				(op in ('greaterThan', 'greaterOrEqualThan') and met.value_num > CAST(md_value AS numeric))				
			)
		)  
	);

END;

$BODY$;

ALTER FUNCTION l3d.search_struct_metadata(hstore, character varying, character varying, character varying)
    OWNER TO i3dhub;




--
-- FUNCTION  l3d.exec_structnodes_searchquery
-- Helper function to execute subqueries
--

DROP FUNCTION IF EXISTS l3d.exec_structnodes_searchquery(text);

CREATE OR REPLACE FUNCTION l3d.exec_structnodes_searchquery(
text)
RETURNS TABLE(id integer, "nodeId" integer,"cache_id" integer, "label" character varying, metadata hstore, tags hstore) 
LANGUAGE 'plpgsql'
AS $BODY$
BEGIN 
    RETURN QUERY EXECUTE $1 ; 
END 
$BODY$;

ALTER FUNCTION l3d.exec_structnodes_searchquery(text)
    OWNER TO i3dhub;

--
-- FUNCTION: l3d.search_structnodes
-- Combines search_struct_metadata and search_struct_label based to fulfil the request
--
-- 

DROP FUNCTION IF EXISTS l3d.search_structnodes(character varying, character varying[], character varying[], character varying[], character varying[], boolean);

CREATE OR REPLACE FUNCTION l3d.search_structnodes(
	cids character varying,
	keys character varying[],
	ops character varying[],
	meta_values character varying[],
	filters character varying[],
	orconditions boolean)
RETURNS TABLE(id integer, "nodeId" integer, "label" character varying, metadata hstore, tags hstore) 
    LANGUAGE 'plpgsql'
    COST 100
    VOLATILE 
    ROWS 1000
AS $BODY$
DECLARE 
/*
select * from l3d.search_structnodes(
(
select array_to_string(array_agg(v), ',') as v
from
(
	select a[1] || '=>' || rtrim(ltrim(replace(array_to_string(string_to_array(a[2], '],[') , ';'), ',' , '-'), '[') , ']') as v
	from
	(
		select  string_to_array(a, '=>') as a
		from
		unnest(string_to_array('04d025567db7f41b-0545aebc-4577-a536/0/154=>[[1,1]];04d025567db7f41b-0545aebc-4577-a536/0/155=>[[29525,13704],[29549,13705],[29573,13706]]', ';')) as a
	) as a
) as a
),'{$InstanceName}', '{equals}', ARRAY['inst_0'],'{nodeId,label,metadata.structMetaData.Time stamp}',false)
*/

	query character varying;
	query_selectPart character varying;
	merg_op character varying;
	metaFilters text[]:= ARRAY[]::text[];
	tagsFilters text[]:= ARRAY[]::text[];
    tkn character varying;
   
    v_cids varchar;
   
   	v_cache_ids hstore;   
    v_offsets hstore;
   
    v_ext_node_id integer;
    v_labels varchar[];
begin	
	 -- in case input data is not formatted as a hstore, but is served directly from the web service's payload 	 	 
	 if cids like '{"%' then
		select array_to_string(array_agg( trim(k_v[1], '"') || '=>' || rtrim(ltrim(replace(replace(k_v[2], '],[', ';'), ',', '-'), '['), ']') ), ',') as v
		into v_cids
		from 
		(
		select string_to_array(rtrim(ltrim(a, '{'), '}'), ':') as k_v
		from 
			unnest(string_to_array(cids, ',"')) as a
		) as a;	 	
	 else
		v_cids := cids;
	 end if;
	
 	-- from input data, collecting caches and offsets data
	select hstore(array_agg(c_id), array_agg(c_id)), hstore(array_agg(c_id), array_agg(offsets))
	into v_cache_ids , v_offsets
	from 
	(
		select coalesce(cac.id::varchar, c.uuid) as c_id, offsets 
		from 
		(
			select c_id[1] uuid, c_id[2]::integer as id0, c_id[3]::integer as id1, offsets
			from 
			(
				select  string_to_array(trim(k_v[1], '"'), '/') as c_id, k_v[2] as offsets
				from
					(select string_to_array(a, '=>') as k_v from unnest(string_to_array(v_cids, ',')) as a) as a
			) as a
		) as c
		left join l3d.cache cac on cac.uuid = c.uuid and cac.ext_cache_id0 = c.id0 and cac.ext_cache_id1 = c.id1 and active = true
	) as a ;


	select coalesce(min(meta_values[i]::integer), -1) into v_ext_node_id
	from
		(select row_number() over() as i, k from unnest(keys) as k) a
	where k ilike 'nodeId'		
	;

	
	select (case when v like '{%}' then v when o = 'contains' then ('{%' || v || '%}') else ('{' || v || '}')  end)::varchar[]
	into v_labels	
	from
	(
		select coalesce(min(replace(meta_values[i], '*', '%')), '{%}') v, coalesce(min(ops[i]), 'contains') o
		from
			(select row_number() over() as i, k from unnest(keys) as k) a
		 where k ilike 'label'
	) a;	

--	raise notice 'v_ext_node_id %, v_labels %', v_ext_node_id, v_labels;

	-- collect the metadata keys and tags names that need to be returned with the results	
	select array_agg(replace(fn, '*', '%')) into metaFilters
	from 
	(
		select names[array_length(names, 1)]  as fn from (select string_to_array(f , '.')  as names from unnest(filters) as f where f ilike '%structMetaData.%') a	
	) a;	

	select array_agg(mf) into tagsFilters
	from 
	(
		select  replace(names[array_length(names, 1)], '*', '%') as mf from (select string_to_array(f , '.')  as names from unnest(filters) as f where f ilike 'tags.%') a
	) a;	
	
--	raise notice 'filters %', filters;

--	RAISE NOTICE 'query %', query_selectPart;
	-- offset handling 

	RETURN QUERY  
	select  t.id, 
			t."nodeId",
			t.label,
			hstore('metadata.array', 
			(
				select array_to_string(array_agg(v), ',')  															
				from
				(
					-- offending characters that could prevent parsing in json
					select regexp_replace(regexp_replace(regexp_replace(a.v, E'\\u0009+', '\\b', 'g'), E'\\u000C+', '\\f', 'g'), E'\\u0009+', '\\t', 'g') as v
					from 
					(
						select regexp_replace(regexp_replace(regexp_replace(a.v, E'\\u000A+', '\\n', 'g'), E'\\u000D+', '\\r', 'g'), E'\\u0000+', '\\0', 'g') as v
						from
						(
							select replace(coalesce(v, '"undefined"'), '@@^^@@', case when v like '{%}' then '' else '"' end) as v 
							from unnest( (t.md_store || hstore('nodeId', t."nodeId"::text))->(filters) ) v										
						) a												
					) a
														
				) a
			)),
			t.tags
	from 
	(
		SELECT  t.id, 
	   			t."nodeId" + split_part(t.offset_root_id, '-', case when t."nodeId" != 0 then 1 else 2 end)::integer as "nodeId",		
				t.label ,
				t.md_store,
				hstore(null) as tags
		FROM 
		(
			select d.id, d."nodeId", d.label, unnest(string_to_array((v_offsets)->(d."cache_id"::text), ';')) as offset_root_id,
				(		 				 
					select 	case when exists(select * from unnest(filters) f where f ilike 'metadata.structMetaData.*') then 
								hstore('metadata.structMetaData.*', '{'|| structMetadata || '}') || (replace(structMetadata, '":"', '"=>"')::hstore)															
							else 
								(replace(structMetadata, '":"', '"=>"')::hstore)														
							end || hstore('Label', '@@^^@@' || d.label || '@@^^@@') || hstore('label', '@@^^@@' || d.label || '@@^^@@')							
					from 
					(
						select coalesce(array_to_string(array_agg('"metadata.structMetaData.' || r.k || '":"' || regexp_replace(r.v, E'[''"\\0\\n]+', ' ', 'g')  || '"'), ',' ), '') structMetadata
					  	from 
					  	(
					  		select mda.key as k, coalesce( '@@^^@@' || replace(mda.value_str, '\', '/')  || '@@^^@@', mda.value_num::text) as v
						  	from l3d.metadata mda 
						  	where 
						  		mda.node_id = d.id and 
						  		mda.key like any(metaFilters) and  
						  		coalesce(array_upper(metaFilters, 1), 0) > 0 	
					  	) r				
					) r
				) as md_store				
			from 
			(
				select nod.id, nod.ext_node_id as "nodeId", nod.cache_id, nod.label
				from node nod
				where 
					not exists( select * from unnest( keys ) k  where k not ilike all( array['nodeId', 'label'] ) )		and
					( v_ext_node_id = -1  or nod.ext_node_id = v_ext_node_id)											and
					( coalesce(array_upper(v_labels, 1), 0) = 0  or nod.label ilike any( v_labels ))					and
					nod.cache_id = any( akeys(v_cids::hstore)::integer[] )
					
				union all
				
				(				
					select nod.id, nod.ext_node_id as "nodeId", nod.cache_id, nod.label				
					from
						node nod,
						l3d.metadata mda,
						( 
							select 	key, 
								 	relation, 
								 	case when value ~ '^\d+(.\d+)?$' then replace(value, ',', '.')::numeric else null end as value_num, 
								 	case when relation = 'contains' then ('%' || replace(value, '/', '\') || '%') else value end as value_str,
									case when relation <> 'equalsAny' then null else value::varchar[] end as value_array						
							from
							(
								select replace( key.name, 'metadata.', '' ) as key, relation.name as relation, replace( trim( value.name, '"' ),'*','%' ) as value
								from
									(select row_number() over() as id, elements as name from unnest(keys) as elements) as key, 
									(select row_number() over() as id, elements as name from unnest(ops) as elements) as relation,
									(select row_number() over() as id, elements as name from unnest(meta_values) as elements) as value
								where
									key.name not ilike all( array['nodeId', 'label'] ) and key.id = relation.id and  key.id = value.id				
							) as a
						) as cnd
					where
						exists( select * from unnest(keys) k  where k not ilike all( array['nodeId', 'label'] ) )			and

						( v_ext_node_id = -1  or nod.ext_node_id = v_ext_node_id)											and
						( coalesce(array_upper(v_labels, 1), 0) = 0  or nod.label ilike any( v_labels ))					and												
						nod.cache_id = any( akeys(v_cids::hstore)::integer[] )												and
						
						mda.node_id  = nod.id 																				and
						mda."key" = cnd.key																					and
						(			
							(
								cnd.relation = 'equalsAny' and coalesce(mda.value_str, mda.value_num::varchar) like any (select v from unnest(cnd.value_array) as v) 
							) or 
							(
								cnd.relation = 'contains' and coalesce(mda.value_str, mda.value_num::varchar) ilike cnd.value_str
							) or			
							(
								cnd.value_num is not null and mda.value_num is not null and
								(
									(cnd.relation in ('equals' , 'lessOrEqualThan', 'greaterOrEqualThan')	and mda.value_num = cnd.value_num) or
									(cnd.relation in ('lessThan', 'lessOrEqualThan') 						and mda.value_num < cnd.value_num) or
									(cnd.relation in ('greaterThan', 'greaterOrEqualThan')  				and mda.value_num > cnd.value_num) 							 
								)
							) or
							(
								mda.value_str is not null and
								(						 
									(cnd.relation in ('equalss', 'lessOrEqualThan', 'greaterOrEqualThan')	and mda.value_str = cnd.value_str)		or
									(cnd.relation in ('lessThan', 'lessOrEqualThan') 						and mda.value_str < cnd.value_str)		or
									(cnd.relation in ('greaterThan', 'greaterOrEqualThan')  				and mda.value_str > cnd.value_str) 		or
									(cnd.relation in('equals')												and mda.value_str ilike cnd.value_str)	
								)					
							)												
						)		
						group by nod.id
						having count(*) = ( select count(*) from unnest(keys) k  where k not ilike all( array['nodeId', 'label'] ) )					
					)
				) d				
			) t
	) t;				
END;

$BODY$;

ALTER FUNCTION l3d.search_structnodes(character varying, character varying[], character varying[], character varying[], character varying[], boolean)
    OWNER TO i3dhub;
